@@ -10,7 +10,7 @@
 (*                                                                     *)
 (***********************************************************************)
 
-(* $Id: pa_o.ml,v 2.5 1999/06/24 12:54:18 ddr Exp $ *)
+(* $Id: pa_o.ml,v 2.14 2000/04/20 10:05:38 ddr Exp $ *)
 
 open Stdpp;
 open Pcaml;
@@ -179,17 +179,6 @@ value infixop4 =
            && symbolchar x 2 :] -> x)
 ;
 
-value test_op_rp =
-  Grammar.Entry.of_parser gram "test_op_rp"
-    (fun strm ->
-       match Stream.npeek 1 strm with
-       [ [("", x)] when is_operator x ->
-           match Stream.npeek 2 strm with
-           [ [_; ("", ")")] -> ()
-           | _ -> raise Stream.Failure ]
-       | _ -> raise Stream.Failure ])
-;
-
 value test_constr_decl =
   Grammar.Entry.of_parser gram "test_constr_decl"
     (fun strm ->
@@ -223,18 +212,22 @@ value test_label_eq =
   Grammar.Entry.of_parser gram "test_label_eq" (test 1)
 ;
 
-value unary_constr_id = ["Some"; "Match_Failure"];
+value constr_arity = ref [("Some", 1); ("Match_Failure", 1)];
 
-value is_constr_expr_not_unary =
+value rec constr_expr_arity =
   fun
-  [ <:expr< $uid:c$ >> -> not (List.exists (fun i -> i = c) unary_constr_id)
-  | _ -> False ]
+  [ <:expr< $uid:c$ >> ->
+      try List.assoc c constr_arity.val with [ Not_found -> 0 ]
+  | <:expr< $uid:_$.$e$ >> -> constr_expr_arity e
+  | _ -> 1 ]
 ;
 
-value is_constr_patt_not_unary =
+value rec constr_patt_arity =
   fun
-  [ <:patt< $uid:c$ >> -> not (List.exists (fun i -> i = c) unary_constr_id)
-  | _ -> False ]
+  [ <:patt< $uid:c$ >> ->
+      try List.assoc c constr_arity.val with [ Not_found -> 0 ]
+  | <:patt< $uid:_$.$p$ >> -> constr_patt_arity p
+  | _ -> 1 ]
 ;
 
 value rec get_seq =
@@ -276,12 +269,13 @@ value rec patt_lid =
 ;
 
 value type_parameter = Grammar.Entry.create gram "type_parameter";
+value fun_def = Grammar.Entry.create gram "fun_def";
 value fun_binding = Grammar.Entry.create gram "fun_binding";
 
 EXTEND
   GLOBAL: interf implem top_phrase use_file sig_item str_item ctyp patt expr
     module_type module_expr let_binding
-    type_parameter fun_binding;
+    type_parameter fun_def fun_binding;
 
   (* Main entry points *)
 
@@ -307,6 +301,8 @@ EXTEND
     [ [ -> MLast.DpNon
       | s = STRING -> MLast.DpStr s
       | i = INT -> MLast.DpInt i
+      | "false" -> MLast.DpIde ["False"]
+      | "true" -> MLast.DpIde ["True"]
       | sl = mod_ident -> MLast.DpIde sl ] ]
   ;
 
@@ -543,12 +539,13 @@ EXTEND
           <:expr< $mkumin loc f e$ >> ]
     | "apply" LEFTA
       [ e1 = expr; e2 = expr ->
-          if is_constr_expr_not_unary e1 then
-            match e2 with
-            [ <:expr< ( $list:el$ ) >> ->
-                List.fold_left (fun e1 e2 -> <:expr< $e1$ $e2$ >>) e1 el
-            | _ -> <:expr< $e1$ $e2$ >> ]
-          else <:expr< $e1$ $e2$ >>
+          match constr_expr_arity e1 with
+          [ 1 -> <:expr< $e1$ $e2$ >>
+          | _ ->
+              match e2 with
+              [ <:expr< ( $list:el$ ) >> ->
+                  List.fold_left (fun e1 e2 -> <:expr< $e1$ $e2$ >>) e1 el
+              | _ -> <:expr< $e1$ $e2$ >> ] ]
       | "assert"; e = expr LEVEL "simple" ->
           let f = <:expr< $str:input_file.val$ >> in
           let bp = <:expr< $int:string_of_int (fst loc)$ >> in
@@ -588,10 +585,12 @@ EXTEND
           <:expr< { $list:lel$ } >>
       | "{"; e = expr LEVEL "simple"; "with"; lel = lbl_expr_list; "}" ->
           <:expr< { ($e$) with $list:lel$ } >>
-      | "("; test_op_rp; op = operator; ")" -> <:expr< $lid:op$ >>
       | "("; ")" -> <:expr< () >>
       | "("; e = expr; ":"; t = ctyp; ")" -> <:expr< ($e$ : $t$) >>
       | "("; e = expr; ")" -> <:expr< $e$ >>
+      | "("; "-"; ")" -> <:expr< $lid:"-"$ >>
+      | "("; "-."; ")" -> <:expr< $lid:"-."$ >>
+      | "("; op = operator; ")" -> <:expr< $lid:op$ >>
       | "begin"; e = expr; "end" -> <:expr< $e$ >>
       | x = LOCATE ->
           let x =
@@ -680,12 +679,23 @@ EXTEND
       [ p1 = patt; "::"; p2 = patt -> <:patt< [$p1$ :: $p2$] >> ]
     | LEFTA
       [ p1 = patt; p2 = patt ->
-          if is_constr_patt_not_unary p1 then
-            match p2 with
-            [ <:patt< ( $list:pl$ ) >> ->
-                List.fold_left (fun p1 p2 -> <:patt< $p1$ $p2$ >>) p1 pl
-            | _ -> <:patt< $p1$ $p2$ >> ]
-          else <:patt< $p1$ $p2$ >> ]
+          match constr_patt_arity p1 with
+          [ 1 -> <:patt< $p1$ $p2$ >>
+          | n ->
+              let p2 =
+                match p2 with
+                [ <:patt< _ >> when n > 1 ->
+                    let pl =
+                      loop n where rec loop n =
+                        if n = 0 then [] else [ <:patt< _ >> :: loop (n - 1) ]
+                    in
+                    <:patt< ( $list:pl$ ) >>
+                | _ -> p2 ]
+              in
+              match p2 with
+              [ <:patt< ( $list:pl$ ) >> ->
+                  List.fold_left (fun p1 p2 -> <:patt< $p1$ $p2$ >>) p1 pl
+              | _ -> <:patt< $p1$ $p2$ >> ] ] ]
     | LEFTA
       [ p1 = patt; "."; p2 = patt -> <:patt< $p1$ . $p2$ >> ]
     | "simple"
@@ -702,10 +712,11 @@ EXTEND
       | "[|"; "|]" -> <:patt< [| |] >>
       | "[|"; pl = patt_semi_list; "|]" -> <:patt< [| $list:pl$ |] >>
       | "{"; lpl = lbl_patt_list; "}" -> <:patt< { $list:lpl$ } >>
-      | "("; test_op_rp; op = operator; ")" -> <:patt< $lid:op$ >>
       | "("; ")" -> <:patt< () >>
       | "("; p = patt; ":"; t = ctyp; ")" -> <:patt< ($p$ : $t$) >>
       | "("; p = patt; ")" -> <:patt< $p$ >>
+      | "("; "-"; ")" -> <:patt< $lid:"-"$ >>
+      | "("; op = operator; ")" -> <:patt< $lid:op$ >>
       | "_" -> <:patt< _ >>
       | x = LOCATE ->
           let x =
@@ -856,9 +867,12 @@ and type_id_list =
         (Stream.Error "lowercase identifier expected") ]
 ;
 
+value class_fun_binding = Grammar.Entry.create gram "class_fun_binding";
+
 EXTEND
   GLOBAL: str_item sig_item expr ctyp
-    class_sig_item class_str_item class_type class_expr;
+    class_sig_item class_str_item class_type class_expr
+    class_fun_binding;
 
   str_item:
     [ [ "class"; cd = LIST1 class_declaration SEP "and" ->
@@ -905,7 +919,7 @@ EXTEND
         ce = class_expr ->
           <:class_expr< let $rec:o2b rf$ $list:lb$ in $ce$ >> ]
     | "apply" NONA
-      [ ce = class_expr; sel = LIST1 (expr LEVEL "simple") ->
+      [ ce = class_expr; sel = LIST1 (expr LEVEL "label") ->
           <:class_expr< $ce$ $list:sel$ >> ]
     | "simple"
       [ "["; ct = ctyp; ","; ctcl = LIST1 ctyp SEP ","; "]";
@@ -1074,3 +1088,86 @@ EXTEND
       | i = LIDENT -> [i] ] ]
   ;
 END;
+
+(* Labels *)
+
+EXTEND
+  GLOBAL: ctyp expr patt fun_def fun_binding class_type class_fun_binding;
+  ctyp: AFTER "arrow"
+    [ NONA
+      [ i = LIDENT; ":"; t = ctyp -> <:ctyp< ~ $i$ : $t$ >>
+      | i = QUESTIONIDENTCOLON; t = ctyp -> <:ctyp< ? $i$ : $t$ >>  ] ]
+  ;
+  ctyp: LEVEL "simple"
+    [ [ "["; OPT "|"; rfl = LIST0 row_field SEP "|"; "]" ->
+          <:ctyp< [| $list:rfl$ |] >>
+      | "["; ">"; OPT "|"; rfl = LIST1 row_field SEP "|"; "]" ->
+          <:ctyp< [| > $list:rfl$ |] >>
+      | "[<"; OPT "|"; (rfl, clos) = row_field_list_dd; "]" ->
+          <:ctyp< [| < $list:rfl$ $dd:clos$ |] >> ] ]
+  ;
+  row_field:
+    [ [ "`"; i = ident -> (i, False, [])
+      | "`"; i = ident; "of"; ao = OPT "&"; l = LIST1 ctyp SEP "&" ->
+          (i, o2b ao, l) ] ]
+  ;
+  row_field_list_dd:
+    [ [ rf = row_field -> ([rf], True)
+      | rf = row_field; "|"; ".." -> ([rf], False)
+      | rf = row_field; "|"; (rfl, dd) = row_field_list_dd ->
+          ([rf :: rfl], dd) ] ]
+  ;
+  expr: LEVEL "expr1"
+    [ [ "fun"; p = labeled_patt; e = fun_def ->
+          <:expr< fun $p$ -> $e$ >> ] ]
+  ;
+  expr: AFTER "apply"
+    [ "label"
+      [ i = TILDEIDENTCOLON; e = expr -> <:expr< ~ $i$ : $e$ >>
+      | i = TILDEIDENT -> <:expr< ~ $i$ >>
+      | i = QUESTIONIDENTCOLON; e = expr -> <:expr< ? $i$ : $e$ >>
+      | i = QUESTIONIDENT -> <:expr< ? $i$ >> ] ]
+  ;
+  expr: LEVEL "simple"
+    [ [ "`"; s = ident -> <:expr< ` $s$ >> ] ]
+  ;
+  fun_def:
+    [ [ p = labeled_patt; e = fun_def ->
+          <:expr< fun $p$ -> $e$ >> ] ]
+  ;
+  fun_binding:
+    [ [ p = labeled_patt; e = fun_binding -> <:expr< fun $p$ -> $e$ >> ] ]
+  ;
+  patt: LEVEL "simple"
+    [ [ "`"; s = ident -> <:patt< ` $s$ >> ] ]
+  ;
+  labeled_patt:
+    [ [ i = TILDEIDENTCOLON; p = patt LEVEL "simple" ->
+          <:patt< ~ $i$ : $p$ >>
+      | i = TILDEIDENT ->
+          <:patt< ~ $i$ >>
+      | i = QUESTIONIDENTCOLON; j = LIDENT ->
+          <:patt< ? $i$ : $lid:j$ >>
+      | i = QUESTIONIDENTCOLON; "("; j = LIDENT; "="; e = expr; ")" ->
+          <:patt< ? $i$ : ( $lid:j$ = $e$ ) >>
+      | i = QUESTIONIDENT ->
+          <:patt< ? $i$ : $lid:i$ >>
+      | "?"; "("; i = LIDENT; "="; e = expr; ")" ->
+          <:patt< ? ( $i$ = $e$ ) >> ] ]
+  ;
+  class_type:
+    [ [ i = LIDENT; ":"; t = ctyp LEVEL "ctyp1"; "->"; ct = class_type ->
+          <:class_type< [ ~ $i$ : $t$ ] -> $ct$ >>
+      | i = QUESTIONIDENTCOLON; t = ctyp LEVEL "ctyp1"; "->";
+        ct = class_type ->
+          <:class_type< [ ? $i$ : $t$ ] -> $ct$ >> ] ]
+  ;
+  class_fun_binding:
+    [ [ p = labeled_patt; cfb = class_fun_binding ->
+          <:class_expr< fun $p$ -> $cfb$ >> ] ]
+  ;
+  ident:
+    [ [ i = LIDENT -> i
+      | i = UIDENT -> i ] ]
+  ;
+END;
