@@ -10,7 +10,7 @@
 (*                                                                     *)
 (***********************************************************************)
 
-(* $Id: pa_op.ml,v 2.1 1999/02/10 18:15:05 ddr Exp $ *)
+(* $Id: pa_op.ml,v 2.2 2000/04/12 22:39:43 ddr Exp $ *)
 
 open Pcaml;
 
@@ -114,7 +114,7 @@ value stream_pattern_component skont ckont =
   | SpNtr loc p e ->
       let e =
         match e with
-        [ <:expr< fun [ $lid:v$ -> $e$ ] >> when v = strm_n -> e
+        [ <:expr< fun [ ($lid:v$ : Stream.t _) -> $e$ ] >> when v = strm_n -> e
         | _ -> <:expr< $e$ $lid:strm_n$ >> ]
       in
       if pattern_eq_expression p skont then
