@@ -10,7 +10,7 @@
 (*                                                                     *)
 (***********************************************************************)
 
-(* $Id: pr_o.ml,v 2.6 1999/06/24 12:54:20 ddr Exp $ *)
+(* $Id: pr_o.ml,v 2.13 2000/04/18 09:07:55 ddr Exp $ *)
 
 open Pretty;
 open Stdpp;
@@ -183,6 +183,16 @@ and variant b (c, tl) _ k =
   | _ -> HVbox [: b; `HOVbox [: `S LR c; `S LR "of"; ctyp_list tl "" k :] :] ]
 ;
 
+value rec row_fields b rfl _ k = listwbws row_field b (S LR "|") rfl "" k
+and row_field b (c, ao, tl) _ k =
+  let c = "`" ^ c in
+  match tl with
+  [ [] -> HVbox [: b; `HOVbox [: `S LR c; k :] :]
+  | _ ->
+      let ao = if ao then [: `S LR "&" :] else [: :] in
+      HVbox [: b; `HOVbox [: `S LR c; `S LR "of"; ao; ctyp_list tl "" k :] :] ]
+;
+
 value rec get_type_args t tl =
   match t with
   [ <:ctyp< $t1$ $t2$ >> -> get_type_args t1 [t2 :: tl]
@@ -239,6 +249,14 @@ simple_ctyp_f.val :=
           [ <:ctyp< $t1$ == $t2$ >> ->
               [: curr t1 "=" [: `S LR "=" :]; `next t2 "" k :]
           | t -> [: `next t "" k :] ]);
+     level (fun x -> HOVbox x)
+       (fun curr next t _ k ->
+          match t with
+          [ <:ctyp< ? $lab$ : $t$ >> ->
+              [: `S LO "?"; `S LR lab; `S RO ":"; `next t ""k :]
+          | <:ctyp< ~ $lab$ : $t$ >> ->
+              [: `S LO (lab ^ ":"); `next t "" k :]
+          | t -> [: `next t "" k :] ]);
      level (fun x -> HOVbox x)
        (fun curr next t _ k ->
           match t with
@@ -272,9 +290,33 @@ simple_ctyp_f.val :=
                     [: labels [: `S LR "{" :] ftl "" [: `S LR "}" :]; k :] :]
           | <:ctyp< [ $list:ctl$ ] >> ->
               [: `BEbox [: `HVbox [: :]; variants [: :] ctl "" [: :]; k :] :]
+          | <:ctyp< [| $list:rfl$ |] >> ->
+              [: `HVbox
+                    [: `HVbox [: :];
+                       row_fields [: `S LR "[" :] rfl "" [: `S LR "]" :];
+                       k :] :]
+          | <:ctyp< [| > $list:rfl$ |] >> ->
+              [: `HVbox
+                    [: `HVbox [: :];
+                       row_fields [: `S LR "[>" :] rfl "" [: `S LR "]" :];
+                       k :] :]
+          | <:ctyp< [| < $list:rfl$ $dd:clos$ $gt:sl$ |] >> ->
+              let k1 = [: `S LR "]" :] in
+              let k1 =
+                match sl with
+                [ [] -> k1
+                | l ->
+                    [: `S LR ">";
+                       list (fun x _ k -> HVbox [: `S LR x; k :]) l "" k1 :] ]
+              in
+              let k1 = if clos then [: `S LR "|"; `S LR ".."; k1 :] else k1 in
+              [: `HVbox
+                    [: `HVbox [: :];
+                       row_fields [: `S LR "["; `S LR "<" :] rfl "" k1; k :] :]
           | (<:ctyp< $_$ -> $_$ >> | <:ctyp< $_$ $_$ >>
           | <:ctyp< $_$ == $_$ >> | <:ctyp< $_$ . $_$ >> |
             <:ctyp< ($list:_$) >> | <:ctyp< $_$ as $_$ >> |
+            <:ctyp< ~ $_$ : $_$ >> | <:ctyp< ? $_$ : $_$ >> |
             MLast.TyXnd _ _ _
             as t) ->
               [: `S LO "("; `ctyp t "" [: `HVbox [: `S RO ")"; k :] :] :]
@@ -282,7 +324,7 @@ simple_ctyp_f.val :=
               [: `S LO "#"; `class_longident id "" k :]
           | MLast.TyObj _ [] False -> [: `S LR "<>"; k :]
           | MLast.TyObj _ ml v ->
-              [: `S LO "<"; meth_list (ml, v) "" [: `S RO ">"; k :] :] ])]
+              [: `S LR "<"; meth_list (ml, v) "" [: `S LR ">"; k :] :] ])]
     (fun t _ k -> not_impl "ctyp" t);
 
 ctyp_f.val :=
@@ -299,6 +341,12 @@ ctyp_f.val :=
           [ <:ctyp< $x$ -> $y$ >> ->
               [: `next x "" [: `S LR "->" :]; curr y "" k :]
           | t -> [: `next t "" k :] ]);
+     level (fun x -> HOVbox x)
+       (fun curr next t _ k ->
+          match t with
+          [ <:ctyp< ? $lab$ : $t$ >> ->
+              [: `S LO "?"; `S LR lab; `S RO ":"; `next t "" k :]
+          | t -> [: `next t "" k :] ]);
      level (fun x -> HOVbox x)
        (fun curr next t _ k ->
           match t with
@@ -317,6 +365,11 @@ value patt1 p dg k = patt1_f.val p dg k;
 value simple_patt_f = ref null_f;
 value simple_patt p dg k = simple_patt_f.val p dg k;
 
+value expr_f = ref null_f;
+value expr1_f = ref null_f;
+value expr e dg k = expr_f.val e dg k;
+value expr1 e dg k = expr1_f.val e dg k;
+
 value rec get_patt_args a al =
   match a with
   [ <:patt< $a1$ $a2$ >> -> get_patt_args a1 [a2 :: al]
@@ -439,6 +492,21 @@ simple_patt_f.val :=
           | <:patt< $chr:c$ >> -> [: `S LR ("'" ^ Char.escaped c ^ "'"); k :]
           | <:patt< $lid:i$ >> -> [: `id_var i; k :]
           | <:patt< $uid:i$ >> -> [: `S LR (conv_con i); k :]
+          | <:patt< ` $i$ >> -> [: `S LR ("`" ^ i); k :]
+          | <:patt< ~ $i$ : $lid:j$ >> when i = j ->
+              [: `S LR ("~" ^ i); k :]
+          | <:patt< ~ $i$ : $p$ >> ->
+              [: `S LO ("~" ^ i ^ ":"); `simple_patt p "" k :]
+          | <:patt< ? $i$ : $lid:j$ >> when i = j ->
+              [: `S LR ("?" ^ i); k :]
+          | <:patt< ? $i$ : $p$ >> ->
+              [: `S LO ("?" ^ i ^ ":"); `simple_patt p "" k :]
+          | <:patt< ? $i$ : ($lid:j$ = $e$) >> when i = j ->
+              [: `S LO "?"; `S LO "("; `S LR j; `S LR "=";
+                 `expr e "" [: `S RO ")"; k :] :]
+          | <:patt< ? $i$ : ($p$ = $e$) >> ->
+              [: `S LO ("?" ^ i ^ ":"); `S LO "(";
+                 `patt p "" [: `S LR "=" :]; `expr e "" [: `S RO ")"; k :] :]
           | <:patt< _ >> -> [: `S LR "_"; k :]
           | <:patt< $_$ $_$ >> | <:patt< ($_$ as $_$) >> |
             <:patt< $_$ | $_$ >> | <:patt< ($list:_$) >> |
@@ -457,6 +525,9 @@ value rec is_irrefut_patt =
       List.for_all (fun (_, p) -> is_irrefut_patt p) fpl
   | <:patt< ($p$ : $_$) >> -> is_irrefut_patt p
   | <:patt< ($list:pl$) >> -> List.for_all is_irrefut_patt pl
+  | <:patt< ? $_$ : $p$ >> -> is_irrefut_patt p
+  | <:patt< ? $_$ : ($p$ = $_$) >> -> is_irrefut_patt p
+  | <:patt< ~ $_$ : $p$ >> -> is_irrefut_patt p
   | _ -> False ]
 ;
 
@@ -474,11 +545,6 @@ value rec expr_fun_args ge =
   | _ -> ([], ge) ]
 ;
 
-value expr_f = ref null_f;
-value expr1_f = ref null_f;
-value expr e dg k = expr_f.val e dg k;
-value expr1 e dg k = expr1_f.val e dg k;
-
 value raise_match_failure (bp, ep) k =
   HOVbox
     [: `S LR "raise"; `S LO "("; `S LR "Match_failure"; `S LO "(";
@@ -588,6 +654,17 @@ value simple_expr =
           | <:expr< $chr:c$ >> -> [: `S LR ("'" ^ Char.escaped c ^ "'"); k :]
           | <:expr< $uid:s$ >> -> [: `S LR (conv_con s); k :]
           | <:expr< $lid:s$ >> -> [: `S LR (var_escaped s); k :]
+          | <:expr< $e$ # $lab$ >> ->
+              [: curr e "" [: :]; `S NO "#"; `label lab; k :]
+          | <:expr< ` $i$ >> -> [: `S LR ("`" ^ i); k :]
+          | <:expr< ~ $i$ : $lid:j$ >> when i = j ->
+              [: `S LR ("~" ^ i); k :]
+          | <:expr< ~ $i$ : $e$ >> ->
+              [: `S LO ("~" ^ i ^ ":"); curr e "" k :]
+          | <:expr< ? $i$ : $lid:j$ >> when i = j ->
+              [: `S LR ("?" ^ i); k :]
+          | <:expr< ? $i$ : $e$ >> ->
+              [: `S LO ("?" ^ i ^ ":"); curr e "" k :]
           | <:expr< [| $list:el$ |] >> ->
               [: `S LR "[|";
                  listws expr (S RO ";") el "" [: `S LR "|]"; k :] :]
@@ -1041,8 +1118,6 @@ expr1_f.val :=
               [: `next e "" k :]
           | <:expr< $e1$ . $e2$ >> ->
               [: curr e1 "" [: :]; `S NO "."; curr e2 "" k :]
-          | MLast.ExSnd _ e lab ->
-              [: curr e "" [: :]; `S NO "#"; `label lab; k :]
           | e -> [: `next e "" k :] ]);
      level (fun x -> HOVbox x)
        (fun curr next e _ k ->
@@ -1433,6 +1508,10 @@ and class_sig_item csf dg k =
       HVbox
         [: `S LR "method"; private_flag pf; `label lab; `S LR ":";
            `ctyp t "" k :]
+  | MLast.CgVal _ lab mf t ->
+      HVbox
+        [: `S LR "val"; mutable_flag mf; `label lab; `S LR ":";
+           `ctyp t "" k :]
   | _ -> HVbox [: `not_impl "class_sig_item" csf; k :] ]
 and class_description b ci _ k =
   HVbox
