@@ -10,7 +10,7 @@
 (*                                                                     *)
 (***********************************************************************)
 
-(* $Id: pr_r.ml,v 2.4 1999/06/24 12:54:22 ddr Exp $ *)
+(* $Id: pr_r.ml,v 2.12 2000/04/18 12:25:41 ddr Exp $ *)
 
 open Pretty;
 
@@ -123,7 +123,7 @@ value has_special_chars v =
 ;
 
 value var_escaped v =
-  if has_special_chars v || is_keyword v then "\\" ^ v
+  if has_special_chars v || is_keyword v then " \\" ^ v
   else v
 ;
 
@@ -146,13 +146,23 @@ and label (f, m, t) k =
 
 value rec ctyp_list tel k = listws ctyp (S LR "and") tel k;
 
-value rec variants b vl k = listwbws variant b (S LR "|") k vl
+value rec variants b vl k = listwbws variant b (S LR "|") vl k
 and variant b (c, tl) k =
   match tl with
   [ [] -> HVbox [: b; `HOVbox [: `S LR c; k :] :]
   | _ -> HVbox [: b; `HOVbox [: `S LR c; `S LR "of"; ctyp_list tl k :] :] ]
 ;
 
+value rec row_fields b rfl k = listwbws row_field b (S LR "|") rfl k
+and row_field b (c, ao, tl) k =
+  let c = "`" ^ c in
+  match tl with
+  [ [] -> HVbox [: b; `HOVbox [: `S LR c; k :] :]
+  | _ ->
+      let ao = if ao then [: `S LR "&" :] else [: :] in
+      HVbox [: b; `HOVbox [: `S LR c; `S LR "of"; ao; ctyp_list tl k :] :] ]
+;
+
 (* *)
 
 value rec class_longident sl k =
@@ -202,6 +212,14 @@ ctyp_f.val :=
           [ <:ctyp< $t1$ $t2$ >> -> [: curr t1 [: :]; `next t2 k :]
           | MLast.TyXnd _ c t -> [: `S LR ("Xnd_" ^ c); `next t k :]
           | t -> [: `next t k :] ]);
+     level (fun x -> HOVbox x)
+       (fun curr next t k ->
+          match t with
+          [ <:ctyp< ? $lab$ : $t$ >> ->
+              [: `S LO "?"; `S LR lab; `S RO ":"; `next t k :]
+          | <:ctyp< ~ $lab$ : $t$ >> ->
+              [: `S LO ("~" ^ lab ^ ":"); `next t k :]
+          | t -> [: `next t k :] ]);
      level (fun x -> HOVbox x)
        (fun curr next t k ->
           match t with
@@ -214,7 +232,7 @@ ctyp_f.val :=
           [ <:ctyp< ($list:tl$) >> ->
               [: `S LO "("; listws ctyp (S LR "*") tl [: `S RO ")"; k :] :]
           | <:ctyp< '$s$ >> -> [: `S LO "'"; `S LR (var_escaped s); k :]
-          | <:ctyp< $lid:s$ >> -> [: `S LR s; k :]
+          | <:ctyp< $lid:s$ >> -> [: `S LR (var_escaped s); k :]
           | <:ctyp< $uid:s$ >> -> [: `S LR s; k :]
           | <:ctyp< _ >> -> [: `S LR "_"; k :]
           | <:ctyp< { $list: ftl$ } >> ->
@@ -222,20 +240,48 @@ ctyp_f.val :=
           | <:ctyp< [ $list:ctl$ ] >> ->
               [: `HVbox
                     [: `HVbox [: :];
-                       variants [: `S LR "[" :] [: `S LR "]" :] ctl; k :] :]
+                       variants [: `S LR "[" :] ctl [: `S LR "]" :]; k :] :]
+          | <:ctyp< [| $list:rfl$ |] >> ->
+              [: `HVbox
+                    [: `HVbox [: :];
+                       row_fields [: `S LR "[|" :] rfl [: `S LR "|]" :];
+                       k :] :]
+          | <:ctyp< [| > $list:rfl$ |] >> ->
+              [: `HVbox
+                    [: `HVbox [: :];
+                       row_fields [: `S LR "[|"; `S LR ">" :] rfl
+                         [: `S LR "|]" :];
+                       k :] :]
+          | <:ctyp< [| < $list:rfl$ $dd:clos$ $gt:sl$ |] >> ->
+              let k1 = [: `S LR "|]" :] in
+              let k1 =
+                match sl with
+                [ [] -> k1
+                | l ->
+                    [: `S LR ">";
+                       list (fun x k -> HVbox [: `S LR x; k :]) l k1 :] ]
+              in
+              let k1 = if clos then [: `S LR "|"; `S LR ".."; k1 :] else k1 in
+              [: `HVbox
+                    [: `HVbox [: :];
+                       row_fields [: `S LR "[|"; `S LR "<" :] rfl k1; k :] :]
           | <:ctyp< $_$ -> $_$ >> | <:ctyp< $_$ $_$ >> |
             <:ctyp< $_$ == $_$ >> | <:ctyp< $_$ . $_$ >> |
-            <:ctyp< $_$ as $_$ >> | MLast.TyXnd _ _ _ ->
+            <:ctyp< $_$ as $_$ >> | <:ctyp< ? $_$ : $_$ >> |
+            <:ctyp< ~ $_$ : $_$ >> | MLast.TyXnd _ _ _ ->
               [: `S LO "("; `ctyp t [: `HVbox [: `S RO ")"; k :] :] :]
           | MLast.TyCls _ id ->
               [: `S LO "#"; `class_longident id k :]
           | MLast.TyObj _ [] False -> [: `S LR "<>"; k :]
           | MLast.TyObj _ ml v ->
-              [: `S LO "<"; meth_list (ml, v) [: `S RO ">"; k :] :] ])]
+              [: `S LR "<"; meth_list (ml, v) [: `S LR ">"; k :] :] ])]
     (fun t k -> not_impl "ctyp" t);
 
 (* patterns *)
 
+value expr_f = ref null_f;
+value expr e k = expr_f.val e k;
+
 value patt_f = ref null_f;
 value patt p k = patt_f.val p k;
 
@@ -316,6 +362,21 @@ patt_f.val :=
           | <:patt< $chr:c$ >> -> [: `S LR ("'" ^ char_escaped c ^ "'"); k :]
           | <:patt< $lid:s$ >> -> [: `S LR (var_escaped s); k :]
           | <:patt< $uid:s$ >> -> [: `S LR s; k :]
+          | <:patt< ` $i$ >> -> [: `S LR ("`" ^ i); k :]
+          | <:patt< ~ $i$ : $lid:j$ >> when i = j ->
+              [: `S LR ("~" ^ i); k :]
+          | <:patt< ~ $i$ : $p$ >> ->
+              [: `S LO ("~" ^ i ^ ":"); curr p k :]
+          | <:patt< ? $i$ : $lid:j$ >> when i = j ->
+              [: `S LR ("?" ^ i); k :]
+          | <:patt< ? $i$ : $p$ >> ->
+              [: `S LO ("?" ^ i ^ ":"); curr p k :]
+          | <:patt< ? $i$ : ($lid:j$ = $e$) >> when i = j ->
+              [: `S LO "?"; `S LO "("; `S LR (var_escaped j); `S LR "=";
+                 `expr e [: `S RO ")"; k :] :]
+          | <:patt< ? $i$ : ($p$ = $e$) >> ->
+              [: `S LO ("?" ^ i ^ ":"); `S LO "(";
+                 `patt p [: `S LR "=" :]; `expr e [: `S RO ")"; k :] :]
           | <:patt< _ >> -> [: `S LR "_"; k :]
           | <:patt< $_$ $_$ >> | <:patt< $_$ .. $_$ >>
           | <:patt< $_$ | $_$ >> | MLast.PaXnd _ _ _ ->
@@ -333,6 +394,9 @@ value rec is_irrefut_patt =
       List.for_all (fun (_, p) -> is_irrefut_patt p) fpl
   | <:patt< ($p$ : $_$) >> -> is_irrefut_patt p
   | <:patt< ($list:pl$) >> -> List.for_all is_irrefut_patt pl
+  | <:patt< ? $_$ : $p$ >> -> is_irrefut_patt p
+  | <:patt< ? $_$ : ($p$ = $_$) >> -> is_irrefut_patt p
+  | <:patt< ~ $_$ : $p$ >> -> is_irrefut_patt p
   | _ -> False ]
 ;
 
@@ -348,12 +412,16 @@ value rec get_defined_ident =
   | <:patt< [| $list:pl$ |] >> -> List.flatten (List.map get_defined_ident pl)
   | <:patt< ($list:pl$) >> -> List.flatten (List.map get_defined_ident pl)
   | <:patt< $uid:_$ >> -> []
+  | <:patt< ` $_$ >> -> []
   | <:patt< $p1$ $p2$ >> -> get_defined_ident p1 @ get_defined_ident p2
   | <:patt< { $list:lpl$ } >> ->
       List.flatten (List.map (fun (lab, p) -> get_defined_ident p) lpl)
   | <:patt< $p1$ | $p2$ >> -> get_defined_ident p1 @ get_defined_ident p2
   | <:patt< $p1$ .. $p2$ >> -> get_defined_ident p1 @ get_defined_ident p2
   | <:patt< ($p$ : $_$) >> -> get_defined_ident p
+  | <:patt< ~ $_$ : $p$ >> -> get_defined_ident p
+  | <:patt< ? $_$ : $p$ >> -> get_defined_ident p
+  | <:patt< ? $_$ : ($p$ = $e$) >> -> get_defined_ident p
   | MLast.PaAnt _ p -> get_defined_ident p
   | MLast.PaXnd _ _ p -> get_defined_ident p ]
 ;
@@ -387,9 +455,6 @@ value rec expr_fun_args ge =
   | _ -> ([], ge) ]
 ;
 
-value expr_f = ref null_f;
-value expr e k = expr_f.val e k;
-
 value rec bind_list b pel k =
   match pel with
   [ [pe] -> let_binding b pe k
@@ -459,7 +524,7 @@ and match_assoc b (p, w, e) k =
 ;
 
 value label lab =
-  S LR lab
+  S LR (var_escaped lab)
 ;
 
 value field_expr (lab, e) k =
@@ -722,7 +787,7 @@ expr_f.val :=
               [: `next e k :]
           | <:expr< $e1$ . $e2$ >> ->
               [: curr e1 [: :]; `S NO "."; curr e2 k :]
-          | MLast.ExSnd _ e lab ->
+          | <:expr< $e$ # $lab$ >> ->
               [: curr e [: :]; `S NO "#"; `label lab; k :]
           | e -> [: `next e k :] ]);
      level (fun x -> HOVbox x)
@@ -744,12 +809,23 @@ expr_f.val :=
           | <:expr< $int:x$ >> ->
               if x.[0] = '-' then [: `S LO "("; `S LR x; `S RO ")"; k :]
               else [: `S LR x; k :]
-          | <:expr< $flo:s$ >> -> [: `S LR s; k :]
+          | <:expr< $flo:x$ >> ->
+              if x.[0] = '-' then [: `S LO "("; `S LR x; `S RO ")"; k :]
+              else [: `S LR x; k :]
           | <:expr< $str:s$ >> ->
               [: `S LR ("\"" ^ String.escaped s ^ "\""); k :]
           | <:expr< $chr:c$ >> -> [: `S LR ("'" ^ char_escaped c ^ "'"); k :]
           | <:expr< $uid:s$ >> -> [: `S LR s; k :]
           | <:expr< $lid:s$ >> -> [: `S LR (var_escaped s); k :]
+          | <:expr< ` $i$ >> -> [: `S LR ("`" ^ i); k :]
+          | <:expr< ~ $i$ : $lid:j$ >> when i = j ->
+              [: `S LR ("~" ^ i); k :]
+          | <:expr< ~ $i$ : $e$ >> ->
+              [: `S LO ("~" ^ i ^ ":"); curr e k :]
+          | <:expr< ? $i$ : $lid:j$ >> when i = j ->
+              [: `S LR ("?" ^ i); k :]
+          | <:expr< ? $i$ : $e$ >> ->
+              [: `S LO ("?" ^ i ^ ":"); curr e k :]
           | (<:expr< [$_$ :: $_$] >> as e) ->
               let (el, c) =
                 make_list e where rec make_list e =
@@ -795,6 +871,7 @@ expr_f.val :=
           | <:expr< ($list:el$) >> ->
               [: `S LO "("; listws expr (S RO ",") el [: `S RO ")"; k :] :]
           | (<:expr< $_$ $_$ >> | <:expr< $uid:_$ $_$ $_$ >> |
+             <:expr< $_$ . $_$ >> | <:expr< $_$ # $_$ >> |
              <:expr< fun [ $list:_$ ] >> |
              <:expr< match $_$ with [ $list:_$ ] >> |
              <:expr< if $_$ then $_$ else $_$ >> |
@@ -821,6 +898,7 @@ value type_list b tdl k =
     [: `HVbox [: :];
        listwbws
          (fun b (tn, tp, te, cl) k ->
+            let tn = var_escaped tn in
             HVbox
               [: `HVbox [: b; `S LR tn; type_params tp [: `S LR "=" :] :];
                  `ctyp te [: :]; list constrain cl k :])
@@ -844,7 +922,7 @@ value value_description s t k =
 value rec mod_ident sl k =
   match sl with
   [ [] -> k
-  | [s] -> [: `S LR s; k :]
+  | [s] -> [: `S LR (var_escaped s); k :]
   | [s :: sl] -> [: `S LR s; `S NO "."; mod_ident sl k :] ]
 ;
 
@@ -1141,7 +1219,7 @@ and class_str_item cf k =
   | MLast.CrIni _ se ->
       HVbox [: `S LR "initializer"; `expr se k :] ]
 and label lab =
-  S LR lab
+  S LR (var_escaped lab)
 and cvalue (lab, mf, e) k =
   HVbox [: flag "mutable" mf; `label lab; `S LR "="; `expr e k :]
 and fun_binding b fb k =
@@ -1157,7 +1235,7 @@ and class_type ct k =
   match ct with
   [ MLast.CtFun _ t ct ->
       HVbox
-        [: `S LO "["; `ctyp t [: `S RO "]"; `S LR "->" :]; `class_type ct k :]
+        [: `S LR "["; `ctyp t [: `S LR "]"; `S LR "->" :]; `class_type ct k :]
   | _ -> class_signature ct k ]
 and class_signature cs k =
   match cs with
@@ -1182,6 +1260,10 @@ and class_sig_item csf k =
       HVbox
         [: `S LR "method"; flag "private" pf; `label lab; `S LR ":";
            `ctyp t k :]
+  | MLast.CgVal _ lab mf t ->
+      HVbox
+        [: `S LR "value"; flag "mutable" mf; `label lab; `S LR ":";
+           `ctyp t k :]
   | _ -> HVbox [: `not_impl "class_sig_item" csf; k :] ]
 and class_description b ci k =
   HVbox
