@@ -10,7 +10,7 @@
 (*                                                                     *)
 (***********************************************************************)
 
-(* $Id: grammar.ml,v 2.2 1999/02/27 11:04:26 ddr Exp $ *)
+(* $Id: grammar.ml,v 2.3 2000/03/28 19:19:14 ddr Exp $ *)
 
 open Stdpp;
 open Gramext;
@@ -291,6 +291,36 @@ value recover parser_of_tree entry nlevn alevn bp a s son strm =
   else do_recover parser_of_tree entry nlevn alevn bp a s son strm
 ;
 
+value std_token_parse =
+  fun
+  [ (p_con, "") ->
+      parser [: `(con, prm) when con = p_con :] -> prm
+  | (p_con, p_prm) ->
+      parser [: `(con, prm) when con = p_con && prm = p_prm :] -> prm ]
+;
+
+value peek_nth n strm =
+  let list = Stream.npeek n strm in
+  loop list n where rec loop list n =
+    match (list, n) with
+    [ ([x :: _], 1) -> Some x
+    | ([_ :: l], n) -> loop l (n - 1)
+    | ([], _) -> None ]
+;
+
+value rec get_token_list entry tokl last_tok tree =
+  match tree with
+  [ Node {node = (Stoken tok as s); son = son; brother = DeadEnd} ->
+      match entry.egram.glexer.Token.tparse tok with
+      [ Some _ ->
+          if tokl = [] then None
+          else Some (List.rev [last_tok :: tokl], last_tok, tree)
+      | None -> get_token_list entry [last_tok :: tokl] tok son ]
+  | _ ->
+      if tokl = [] then None
+      else Some (List.rev [last_tok :: tokl], last_tok, tree) ]
+;
+
 value rec parser_of_tree entry nlevn alevn =
   fun
   [ DeadEnd -> parser []
@@ -301,35 +331,95 @@ value rec parser_of_tree entry nlevn alevn =
       let p2 = parser_of_tree entry nlevn alevn bro in
       parser [ [: a = entry.estart alevn :] -> app act a | [: a = p2 :] -> a ]
   | Node {node = s; son = son; brother = DeadEnd} ->
-      let ps = parser_of_symbol entry nlevn s in
-      let p1 = parser_of_tree entry nlevn alevn son in
-      parser bp
-        [: a = ps;
-           act =
-             parser
-             [ [: a = p1 :] -> a
-             | [: a =
-                    recover parser_of_tree entry nlevn alevn bp a s son :] ->
-                 a
-             | [: :] ->
-                 raise (Stream.Error (tree_failed entry a s son)) ] :] ->
-          app act a
+      let tokl =
+        match s with
+        [ Stoken tok when entry.egram.glexer.Token.tparse tok = None ->
+            get_token_list entry [] tok son
+        | _ -> None ]
+      in
+      match tokl with
+      [ None ->
+          let ps = parser_of_symbol entry nlevn s in
+          let p1 = parser_of_tree entry nlevn alevn son in
+          let p1 = parser_cont p1 entry nlevn alevn s son in
+          parser bp [: a = ps; act = p1 bp a :] -> app act a
+      | Some (tokl, last_tok, son) ->
+          let p1 = parser_of_tree entry nlevn alevn son in
+          let p1 = parser_cont p1 entry nlevn alevn (Stoken last_tok) son in
+          parser_of_token_list p1 tokl ]
   | Node {node = s; son = son; brother = bro} ->
-      let ps = parser_of_symbol entry nlevn s in
-      let p1 = parser_of_tree entry nlevn alevn son in
-      let p2 = parser_of_tree entry nlevn alevn bro in
-      parser bp
-      [ [: a = ps;
-           act =
-             parser
-             [ [: a = p1 :] -> a
-             | [: a =
-                    recover parser_of_tree entry nlevn alevn bp a s son :] ->
-                 a
-             | [: :] ->
-                 raise (Stream.Error (tree_failed entry a s son)) ] :] ->
+      let tokl =
+        match s with
+        [ Stoken tok when entry.egram.glexer.Token.tparse tok = None ->
+            get_token_list entry [] tok son
+        | _ -> None ]
+      in
+      match tokl with
+      [ None ->
+          let ps = parser_of_symbol entry nlevn s in
+          let p1 = parser_of_tree entry nlevn alevn son in
+          let p1 = parser_cont p1 entry nlevn alevn s son in
+          let p2 = parser_of_tree entry nlevn alevn bro in
+          parser bp
+          [ [: a = ps; act = p1 bp a :] -> app act a
+          | [: a = p2 :] -> a ]
+      | Some (tokl, last_tok, son) ->
+          let p1 = parser_of_tree entry nlevn alevn son in
+          let p1 = parser_cont p1 entry nlevn alevn (Stoken last_tok) son in
+          let p1 = parser_of_token_list p1 tokl in
+          let p2 = parser_of_tree entry nlevn alevn bro in
+          parser
+          [ [: a = p1 :] -> a
+          | [: a = p2 :] -> a ] ] ]
+and parser_cont p1 entry nlevn alevn s son bp a =
+  parser
+  [ [: a = p1 :] -> a
+  | [: a = recover parser_of_tree entry nlevn alevn bp a s son :] -> a
+  | [: :] -> raise (Stream.Error (tree_failed entry a s son)) ]
+and parser_of_token_list p1 tokl =
+  loop 1 tokl where rec loop n =
+    fun
+    [ [(p_con, "")] ->
+        let ps (strm__ : Stream.t _) =
+          match peek_nth n strm__ with
+          [ Some (con, prm) when con = p_con ->
+              do for i = 1 to n do Stream.junk strm__; done; return
+              Obj.repr prm
+          | _ -> raise Stream.Failure ]
+        in
+        parser bp [: a = ps; act = p1 bp a :] -> app act a
+    | [(p_con, p_prm)] ->
+        let ps (strm__ : Stream.t _) =
+          match peek_nth n strm__ with
+          [ Some (con, prm) when con = p_con && prm = p_prm ->
+              do for i = 1 to n do Stream.junk strm__; done; return
+              Obj.repr prm
+          | _ -> raise Stream.Failure ]
+        in
+        parser bp [: a = ps; act = p1 bp a :] -> app act a
+    | [(p_con, "") :: tokl] ->
+        let ps (strm__ : Stream.t _) =
+          match peek_nth n strm__ with
+          [ Some (con, prm) when con = p_con -> prm
+          | _ -> raise Stream.Failure ]
+        in
+        let p1 = loop (n + 1) tokl in
+        fun (strm__ : Stream.t _) ->
+          let a = ps strm__ in
+          let act = p1 strm__ in
           app act a
-      | [: a = p2 :] -> a ] ]
+    | [(p_con, p_prm) :: tokl] ->
+        let ps (strm__ : Stream.t _) =
+          match peek_nth n strm__ with
+          [ Some (con, prm) when con = p_con && prm = p_prm -> prm
+          | _ -> raise Stream.Failure ]
+        in
+        let p1 = loop (n + 1) tokl in
+        fun (strm__ : Stream.t _) ->
+          let a = ps strm__ in
+          let act = p1 strm__ in
+          app act a
+    | [] -> assert False ]
 and parser_of_symbol entry nlevn =
   fun
   [ Slist0 s ->
@@ -387,10 +477,10 @@ and parser_of_symbol entry nlevn =
   | Sself -> parser [: a = entry.estart 0 :] -> a
   | Snext -> parser [: a = entry.estart nlevn :] -> a
   | Stoken tok ->
-      (Obj.magic
-         (entry.egram.glexer.Token.tparse tok :
-          Stream.t Token.t -> string) :
-       Stream.t Token.t -> Obj.t) ]
+      match entry.egram.glexer.Token.tparse tok with
+      [ Some f -> (Obj.magic f : Stream.t Token.t -> Obj.t)
+      | None ->
+          (Obj.magic (std_token_parse tok) : Stream.t Token.t -> Obj.t) ] ]
 ;
 
 value rec continue_parser_of_levels entry clevn =
