@@ -10,7 +10,7 @@
 (*                                                                     *)
 (***********************************************************************)
 
-(* $Id: plexer.ml,v 2.1 1998/12/04 18:29:12 ddr Exp $ *)
+(* $Id: plexer.ml,v 2.5 2000/04/13 18:23:44 ddr Exp $ *)
 
 open Stdpp;
 open Token;
@@ -128,7 +128,7 @@ value escape_char =
 
 value rec skip_spaces =
   parser
-  [ [: `' ' | '\n' | '\r' | '\t' | '\026' | '\012'; s :] -> skip_spaces s
+  [ [: `' ' | '\010' | '\013' | '\t' | '\026' | '\012'; s :] -> skip_spaces s
   | [: :] -> () ]
 ;
 
@@ -138,9 +138,9 @@ value rec skip_indent =
   | [: :] -> () ]
 ;
 
-value skip_opt_newline =
+value skip_opt_linefeed =
   parser
-  [ [: `'\n' :] -> ()
+  [ [: `'\010' :] -> ()
   | [: :] -> () ]
 ;
 
@@ -171,12 +171,36 @@ value next_token_fun find_id_kwd find_spe_kwd =
         | _ -> keyword_or_error (bp, ep) "'" ]
     | [: `'"'; s :] -> ("STRING", string bp 0 s)
     | [: `'$'; s :] -> locate_or_antiquot bp 0 s
-    | [: `('!' | '?' | '~' | '=' | '@' | '^' | '&' | '+' | '-' | '*' | '/' |
+    | [: `('!' | '=' | '@' | '^' | '&' | '+' | '-' | '*' | '/' |
            '%' as
            c);
          s :] ->
         let id = get_buff (ident2 (store 0 c) s) in
         keyword_or_error (bp, Stream.count s) id
+    | [: `('~' as c);
+         r =
+           parser
+           [ [: `('a'..'z' as c); len = ident (store 0 c);
+                t =
+                  parser
+                    [ [: `':' :] -> "TILDEIDENTCOLON"
+                    | [: :] -> "TILDEIDENT" ] :] ->
+               (t, get_buff len)
+           | [: s :] ->
+               let id = get_buff (ident2 (store 0 c) s) in
+               keyword_or_error (bp, Stream.count s) id ] :] -> r
+    | [: `('?' as c);
+         r =
+           parser
+           [ [: `('a'..'z' as c); len = ident (store 0 c);
+                t =
+                  parser
+                    [ [: `':' :] -> "QUESTIONIDENTCOLON"
+                    | [: :] -> "QUESTIONIDENT" ] :] ->
+               (t, get_buff len)
+           | [: s :] ->
+               let id = get_buff (ident2 (store 0 c) s) in
+               keyword_or_error (bp, Stream.count s) id ] :] -> r
     | [: `'<'; s :] -> less bp s
     | [: `(':' as c1);
          len =
@@ -222,8 +246,8 @@ value next_token_fun find_id_kwd find_spe_kwd =
     | [: `'\\';
          len =
            parser
-           [ [: `'\n'; _ = skip_indent :] -> len
-           | [: `'\r'; _ = skip_opt_newline; _ = skip_indent :] -> len
+           [ [: `'\010'; _ = skip_indent :] -> len
+           | [: `'\013'; _ = skip_opt_linefeed; _ = skip_indent :] -> len
            | [: c = escape_string :] -> store len c
            | [: `c :] -> store (store len '\\') c ];
          s :] ->
@@ -308,7 +332,7 @@ value next_token_fun find_id_kwd find_spe_kwd =
   in
   let rec next_token_loc =
     parser bp
-    [ [: `' ' | '\n' | '\r' | '\t' | '\026' | '\012'; s :] -> next_token_loc s
+    [ [: `' ' | '\010' | '\013' | '\t' | '\026' | '\012'; s :] -> next_token_loc s
     | [: `'('; s :] -> maybe_comment bp s
     | [: `'#'; _ = spaces_tabs; t = linenum bp :] -> t
     | [: tok = next_token :] ep -> (tok, (bp, ep))
@@ -344,7 +368,7 @@ value next_token_fun find_id_kwd find_spe_kwd =
     | [: :] -> () ]
   and any_to_nl =
     parser
-    [ [: `'\r' | '\n' :] -> ()
+    [ [: `'\013' | '\010' :] -> ()
     | [: `_; s :] -> any_to_nl s
     | [: :] -> () ]
   in
@@ -451,8 +475,9 @@ value using_token kwd_table (p_con, p_prm) =
           else
             raise (Token.Error ("\
 the token \"" ^ p_prm ^ "\" does not respect Plexer rules")) ]
-  | "LIDENT" | "UIDENT" | "INT"| "FLOAT" | "CHAR" | "STRING" | "QUOTATION"
-  | "ANTIQUOT" | "LOCATE" | "EOI" -> ()
+  | "LIDENT" | "UIDENT" | "TILDEIDENT" | "TILDEIDENTCOLON"
+  | "QUESTIONIDENT" | "QUESTIONIDENTCOLON" | "INT" | "FLOAT"
+  | "CHAR" | "STRING" | "QUOTATION" | "ANTIQUOT" | "LOCATE" | "EOI" -> ()
   | _ ->
       raise (Token.Error ("\
 the constructor \"" ^ p_con ^ "\" is not recognized by Plexer")) ]
@@ -501,12 +526,12 @@ value after_colon e =
 value tparse =
   fun
   [ ("ANTIQUOT", p_prm) ->
-      parser [: `("ANTIQUOT", prm) when eq_before_colon p_prm prm :] ->
-        after_colon prm
-  | (p_con, "") ->
-      parser [: `(con, prm) when con = p_con :] -> prm
-  | (p_con, p_prm) ->
-      parser [: `(con, prm) when con = p_con && prm = p_prm :] -> prm ]
+      let p =
+        parser [: `("ANTIQUOT", prm) when eq_before_colon p_prm prm :] ->
+          after_colon prm
+      in
+      Some p
+  | _ -> None ]
 ;
 
 value make () =
