@@ -10,7 +10,7 @@
 (*                                                                     *)
 (***********************************************************************)
 
-(* $Id: token.ml,v 2.1 1998/12/04 18:29:12 ddr Exp $ *)
+(* $Id: token.ml,v 2.2 2000/03/28 19:19:16 ddr Exp $ *)
 
 type t = (string * string);
 type pattern = (string * string);
@@ -23,6 +23,6 @@ type lexer =
  { func : Stream.t char -> (Stream.t t * location_function);
    using : pattern -> unit;
    removing : pattern -> unit;
-   tparse : pattern -> Stream.t t -> string;
+   tparse : pattern -> option (Stream.t t -> string);
    text : pattern -> string }
 ;
