@@ -10,7 +10,7 @@
 (*                                                                     *)
 (***********************************************************************)
 
-(* $Id: pa_r.ml,v 2.3 1999/06/24 12:54:23 ddr Exp $ *)
+(* $Id: pa_r.ml,v 2.5 2000/04/18 12:25:43 ddr Exp $ *)
 
 open Stdpp;
 open Pcaml;
@@ -81,11 +81,12 @@ value neg s = string_of_int (- int_of_string s);
 
 value type_parameter = Grammar.Entry.create gram "type_parameter";
 value fun_binding = Grammar.Entry.create gram "fun_binding";
+value ipatt = Grammar.Entry.create gram "ipatt";
 
 EXTEND
   GLOBAL: interf implem top_phrase use_file sig_item str_item ctyp patt expr
     module_type module_expr let_binding
-    type_parameter fun_binding;
+    type_parameter fun_binding ipatt;
 
   interf:
     [ [ st = LIST0 [ s = sig_item; ";" -> (s, loc) ]; EOI -> st ] ]
@@ -544,7 +545,7 @@ EXTEND
     [ [ "="; ce = class_expr -> ce
       | ":"; ct = class_type; "="; ce = class_expr ->
           <:class_expr< ($ce$ : $ct$) >>
-      | p = patt LEVEL "simple"; cfb = class_fun_binding ->
+      | p = ipatt; cfb = class_fun_binding ->
           <:class_expr< fun $p$ -> $cfb$ >> ] ]
   ;
   class_type_parameters:
@@ -722,3 +723,63 @@ EXTEND
       | i = LIDENT -> [i] ] ]
   ;
 END;
+
+(* Labels *)
+
+EXTEND
+  GLOBAL: ctyp ipatt patt expr;
+  ctyp: AFTER "arrow"
+    [ NONA
+      [ i = TILDEIDENTCOLON; t = ctyp -> <:ctyp< ~ $i$ : $t$ >>
+      | i = QUESTIONIDENTCOLON; t = ctyp -> <:ctyp< ? $i$ : $t$ >>  ] ]
+  ;
+  ctyp: LEVEL "simple"
+    [ [ "[|"; rfl = LIST0 row_field SEP "|"; "|]" ->
+          <:ctyp< [| $list:rfl$ |] >>
+      | "[|"; ">"; rfl = LIST1 row_field SEP "|"; "|]" ->
+          <:ctyp< [| > $list:rfl$ |] >>
+      | "[|"; "<"; (rfl, clos) = row_field_list_dd; "|]" ->
+          <:ctyp< [| < $list:rfl$ $dd:clos$ |] >> ] ]
+  ;
+  row_field:
+    [ [ "`"; i = ident -> (i, False, [])
+      | "`"; i = ident; "of"; ao = OPT "&"; l = LIST1 ctyp SEP "&" ->
+          (i, o2b ao, l) ] ]
+  ;
+  row_field_list_dd:
+    [ [ rf = row_field -> ([rf], True)
+      | rf = row_field; "|"; ".." -> ([rf], False)
+      | rf = row_field; "|"; (rfl, dd) = row_field_list_dd ->
+          ([rf :: rfl], dd) ] ]
+  ;
+  patt: LEVEL "simple"
+    [ [ "`"; s = ident -> <:patt< ` $s$ >> ] ]
+  ;
+  ipatt:
+    [ [ i = TILDEIDENTCOLON; p = ipatt ->
+          <:patt< ~ $i$ : $p$ >>
+      | i = TILDEIDENT ->
+          <:patt< ~ $i$ >>
+      | i = QUESTIONIDENTCOLON; j = LIDENT ->
+          <:patt< ? $i$ : $lid:j$ >>
+      | i = QUESTIONIDENTCOLON; "("; j = LIDENT; "="; e = expr; ")" ->
+          <:patt< ? $i$ : ( $lid:j$ = $e$ ) >>
+      | i = QUESTIONIDENT ->
+          <:patt< ? $i$ : $lid:i$ >>
+      | "?"; "("; i = LIDENT; "="; e = expr; ")" ->
+          <:patt< ? ( $i$ = $e$ ) >> ] ]
+  ;
+  expr: AFTER "apply"
+    [ "label"
+      [ i = TILDEIDENTCOLON; e = expr -> <:expr< ~ $i$ : $e$ >>
+      | i = TILDEIDENT -> <:expr< ~ $i$ >>
+      | i = QUESTIONIDENTCOLON; e = expr -> <:expr< ? $i$ : $e$ >>
+      | i = QUESTIONIDENT -> <:expr< ? $i$ >> ] ]
+  ;
+  expr: LEVEL "simple"
+    [ [ "`"; s = ident -> <:expr< ` $s$ >> ] ]
+  ;
+  ident:
+    [ [ i = LIDENT -> i | i = UIDENT -> i ] ]
+  ;
+END;
