@@ -10,7 +10,7 @@
 (*                                                                     *)
 (***********************************************************************)
 
-(* $Id: q_MLast.ml,v 2.1 1999/09/06 09:58:43 ddr Exp $ *)
+(* $Id: q_MLast.ml,v 2.5 2000/04/17 21:10:28 ddr Exp $ *)
 
 value gram = Grammar.create (Plexer.make ());
 
@@ -267,6 +267,15 @@ EXTEND
       [ f = [ op = "-" -> op | op = "-." -> op ]; e = SELF -> mkumin f e ]
     | "apply" LEFTA
       [ e1 = SELF; e2 = SELF -> Node "ExApp" [e1; e2] ]
+    | "label" NONA
+      [ lab = TILDEIDENTCOLON; e = SELF -> Node "ExLab" [Str lab; e]
+      | lab = TILDEIDENT -> Node "ExLab" [Str lab; Node "ExLid" [Str lab]]
+      | lab = QUESTIONIDENTCOLON; e = SELF -> Node "ExOlb" [Str lab; e]
+      | lab = QUESTIONIDENT -> Node "ExOlb" [Str lab; Node "ExLid" [Str lab]]
+      | "~"; a = anti_; ":"; e = SELF -> Node "ExLab" [a; e]
+      | "~"; a = anti_ -> Node "ExLab" [a; Node "ExLid" [a]]
+      | "?"; a = anti_; ":"; e = SELF -> Node "ExOlb" [a; e]
+      | "?"; a = anti_ -> Node "ExOlb" [a; Node "ExLid" [a]] ]
     | "." LEFTA
       [ e1 = SELF; "."; "("; e2 = SELF; ")" -> Node "ExAre" [e1; e2]
       | e1 = SELF; "."; "["; e2 = SELF; "]" -> Node "ExSte" [e1; e2]
@@ -281,6 +290,7 @@ EXTEND
       | s = CHAR -> Node "ExChr" [Chr s.[0]]
       | s = UIDENT -> Node "ExUid" [Str s]
       | s = LIDENT -> Node "ExLid" [Str s]
+      | "`"; s = ident -> Node "ExVrn" [s]
       | a = anti_int -> Node "ExInt" [a]
       | a = anti_flo -> Node "ExFlo" [a]
       | a = anti_str -> Node "ExStr" [a]
@@ -346,6 +356,19 @@ EXTEND
     | [ p1 = SELF; ".."; p2 = SELF -> Node "PaRng" [p1; p2] ]
     | [ p1 = SELF; p2 = SELF -> Node "PaApp" [p1; p2] ]
     | [ p1 = SELF; "."; p2 = SELF -> Node "PaAcc" [p1; p2] ]
+    | NONA
+      [ "~"; i = lident; ":"; p = SELF ->
+          Node "PaLab" [i; p]
+      | "~"; i = lident ->
+          Node "PaLab" [i; Node "PaLid" [i]]
+      | "?"; i = lident; ":"; p = SELF ->
+          Node "PaOlb" [i; p; Option None]
+      | "?"; i = lident; ":"; "("; p = SELF; "="; e = expr; ")" ->
+          Node "PaOlb" [i; p; Option (Some e)]
+      | "?"; "("; j = lident; "="; e = expr; ")" ->
+          Node "PaOlb" [j; Node "PaLid" [j]; Option (Some e)]
+      | "?"; i = lident ->
+          Node "PaOlb" [i; Node "PaLid" [i]; Option None] ]
     | "simple"
       [ v = LIDENT -> Node "PaLid" [Str v]
       | v = UIDENT -> Node "PaUid" [Str v]
@@ -353,6 +376,7 @@ EXTEND
       | "-"; s = INT -> Node "PaInt" [Str (neg s)]
       | s = STRING -> Node "PaStr" [Str s]
       | s = CHAR -> Node "PaChr" [Chr s.[0]]
+      | "`"; s = ident -> Node "PaVrn" [s]
       | a = anti_lid -> Node "PaLid" [a]
       | a = anti_uid -> Node "PaUid" [a]
       | a = anti_int -> Node "PaInt" [a]
@@ -420,6 +444,10 @@ EXTEND
       [ t1 = SELF; "as"; t2 = SELF -> Node "TyAli" [t1; t2] ]
     | RIGHTA
       [ t1 = SELF; "->"; t2 = SELF -> Node "TyArr" [t1; t2] ]
+    | NONA
+      [ a = TILDEIDENTCOLON; ":"; t = SELF -> Node "TyLab" [Str a; t]
+      | "~"; a = anti_; ":"; t = SELF -> Node "TyLab" [a; t]
+      | "?"; a = lident; ":"; t = SELF -> Node "TyOlb" [a; t] ]
     | LEFTA
       [ t1 = SELF; t2 = SELF -> Node "TyApp" [t1; t2] ]
     | LEFTA
@@ -439,7 +467,36 @@ EXTEND
       | "["; cdl = SLIST0 constructor_declaration SEP "|"; "]" ->
           Node "TySum" [cdl]
       | "{"; ldl = SLIST1 label_declaration SEP ";"; "}" ->
-          Node "TyRec" [ldl] ] ]
+          Node "TyRec" [ldl]
+      | "[|"; rfl = SLIST0 row_field SEP "|"; "|]" ->
+          Node "TyVrn" [rfl; Option None ]
+      | "[|"; ">"; rfl = SLIST1 row_field SEP "|"; "|]" ->
+          Node "TyVrn" [rfl; Option (Some (Option None))]
+      | "[|"; "<"; (rfl, clos) = row_field_list_dd; sl = opt_tag_list; "|]" ->
+          Node "TyVrn"
+            [rfl; Option (Some (Option (Some (Tuple [clos; sl]))))] ] ]
+  ;
+  row_field:
+    [ [ "`"; i = lident -> Tuple [i; Bool True; List []]
+      | "`"; i = lident; "of"; oa = OPT "&"; l = SLIST1 ctyp SEP "&" ->
+          Tuple [i; Bool (oa <> None); l] ] ]
+  ;
+  row_field_list_dd:
+    [ [ a = anti_list; clos = clos -> (a, clos)
+      | a = anti_list -> (a, Bool True)
+      | rf = row_field; clos = clos -> (rf, clos)
+      | rf = row_field -> (rf, Bool True)
+      | rf = row_field; "|"; (rfl, clos) = row_field_list_dd ->
+          (Cons rf rfl, clos) ] ]
+  ;
+  clos:
+    [ [ a = anti_dd -> a
+      | "|"; ".." -> Bool False ] ]
+  ;
+  opt_tag_list:
+    [ [ a = anti_gt -> a
+      | ">"; sl = LIST1 lident -> List sl
+      | -> List [] ] ]
   ;
   constructor_declaration:
     [ [ ci = uident; "of"; cal = SLIST1 ctyp SEP "and" -> Tuple [ci; cal]
@@ -500,12 +557,18 @@ EXTEND
   anti_chr:
     [ [ a = ANTIQUOT "chr" -> antiquot "chr" loc a ] ]
   ;
+  anti_dd:
+    [ [ a = ANTIQUOT "dd" -> antiquot "dd" loc a ] ]
+  ;
   anti_exp:
     [ [ a = ANTIQUOT "exp" -> antiquot "exp" loc a ] ]
   ;
   anti_flo:
     [ [ a = ANTIQUOT "flo" -> antiquot "flo" loc a ] ]
   ;
+  anti_gt:
+    [ [ a = ANTIQUOT "gt" -> antiquot "gt" loc a ] ]
+  ;
   anti_int:
     [ [ a = ANTIQUOT "int" -> antiquot "int" loc a ] ]
   ;
