@@ -26,7 +26,6 @@ let get_tag x =
 ;;
 
 let error loc str = raise_with_loc loc (Failure str);;
-let loc_of_node v : MLast.loc = Obj.magic (Obj.field (Obj.repr v) 0);;
 
 let mkloc (bp, ep) =
   {Location.loc_start = bp; Location.loc_end = ep; Location.loc_ghost = false}
@@ -118,14 +117,14 @@ let rec ctyp_long_id =
   | TyUid (_, s) -> false, lident s
   | TyLid (_, s) -> false, lident s
   | TyCls (loc, sl) -> true, long_id_of_string_list loc sl
-  | t -> error (loc_of_node t) "incorrect type"
+  | t -> error (loc_of_ctyp t) "incorrect type"
 ;;
 
 let rec ctyp =
   function
     TyAcc (loc, _, _) as f ->
       let (is_cls, li) = ctyp_long_id f in
-      if is_cls then mktyp loc (Ptyp_class (li, []))
+      if is_cls then mktyp loc (Ptyp_class (li, [], []))
       else mktyp loc (Ptyp_constr (li, []))
   | TyAli (loc, t1, t2) ->
       let (t, i) =
@@ -139,19 +138,35 @@ let rec ctyp =
   | TyApp (loc, _, _) as f ->
       let (f, al) = ctyp_fa [] f in
       let (is_cls, li) = ctyp_long_id f in
-      if is_cls then mktyp loc (Ptyp_class (li, List.map ctyp al))
+      if is_cls then mktyp loc (Ptyp_class (li, List.map ctyp al, []))
       else mktyp loc (Ptyp_constr (li, List.map ctyp al))
-  | TyArr (loc, t1, t2) -> mktyp loc (Ptyp_arrow (ctyp t1, ctyp t2))
+  | TyArr (loc, TyLab (loc1, lab, t1), t2) ->
+      mktyp loc (Ptyp_arrow (lab, ctyp t1, ctyp t2))
+  | TyArr (loc, TyOlb (loc1, lab, t1), t2) ->
+      let t1 = TyApp (loc1, TyLid (loc1, "option"), t1) in
+      mktyp loc (Ptyp_arrow (("?" ^ lab), ctyp t1, ctyp t2))
+  | TyArr (loc, t1, t2) -> mktyp loc (Ptyp_arrow ("", ctyp t1, ctyp t2))
   | TyObj (loc, fl, v) -> mktyp loc (Ptyp_object (meth_list loc fl v))
   | TyCls (loc, id) ->
-      mktyp loc (Ptyp_class (long_id_of_string_list loc id, []))
+      mktyp loc (Ptyp_class (long_id_of_string_list loc id, [], []))
+  | TyLab (loc, _, _) -> error loc "labeled type not allowed here"
   | TyLid (loc, s) -> mktyp loc (Ptyp_constr (lident s, []))
   | TyMan (loc, _, _) -> error loc "type manifest not allowed here"
+  | TyOlb (loc, lab, _) -> error loc "labeled type not allowed here"
   | TyQuo (loc, s) -> mktyp loc (Ptyp_var (shd s))
   | TyRec (loc, _) -> error loc "record type not allowed here"
   | TySum (loc, _) -> error loc "sum type not allowed here"
   | TyTup (loc, tl) -> mktyp loc (Ptyp_tuple (List.map ctyp tl))
   | TyUid (loc, s) -> mktyp loc (Ptyp_constr (lident s, []))
+  | TyVrn (loc, catl, ool) ->
+      let catl = List.map (fun (c, a, t) -> c, a, List.map ctyp t) catl in
+      let (clos, sl) =
+        match ool with
+          None -> true, List.map (fun (c, _, _) -> c) catl
+        | Some None -> false, List.map (fun (c, _, _) -> c) catl
+        | Some (Some (clos, sl)) -> clos, sl
+      in
+      mktyp loc (Ptyp_variant (catl, clos, sl))
   | TyXnd (loc, c, _) ->
       error loc ("type \"" ^ c ^ "\" (extension) not allowed here")
 and meth_list loc fl v =
@@ -185,11 +200,17 @@ let type_decl tl cl =
           TyQuo (_, s) -> if List.mem s tl then Some (ctyp t) else None
         | _ -> Some (ctyp t)
       in
-      mktype (loc_of_node t) tl cl Ptype_abstract m
+      mktype (loc_of_ctyp t) tl cl Ptype_abstract m
 ;;
 
 let mkvalue_desc t p = {pval_type = ctyp t; pval_prim = List.map shd p};;
 
+let option f =
+  function
+    Some x -> Some (f x)
+  | None -> None
+;;
+
 let rec same_type_expr ct ce =
   match ct, ce with
     TyLid (_, s1), ExLid (_, s2) -> s1 = s2
@@ -227,14 +248,14 @@ let rec module_type_long_id =
       Lapply (module_type_long_id m1, module_type_long_id m2)
   | MtLid (_, s) -> lident s
   | MtUid (_, s) -> lident s
-  | t -> error (loc_of_node t) "bad module type long ident"
+  | t -> error (loc_of_module_type t) "bad module type long ident"
 ;;
 
 let rec module_expr_long_id =
   function
     MeAcc (_, m, MeUid (_, s)) -> ldot (module_expr_long_id m) s
   | MeUid (_, s) -> lident s
-  | t -> error (loc_of_node t) "bad module expr long ident"
+  | t -> error (loc_of_module_expr t) "bad module expr long ident"
 ;;
 
 let mkwithc =
@@ -277,7 +298,7 @@ let rec patt_label_long_id =
   | PaAcc (_, m, PaUid (_, s)) -> ldot (patt_label_long_id m) s
   | PaUid (_, s) -> lident s
   | PaLid (_, s) -> lident (conv_lab s)
-  | p -> error (loc_of_node p) "bad label"
+  | p -> error (loc_of_patt p) "bad label"
 ;;
 
 let rec patt =
@@ -289,9 +310,9 @@ let rec patt =
             begin match p2 with
               PaUid (_, s) ->
                 Ppat_construct (mkli (conv_con s) (i :: il), None, true)
-            | _ -> error (loc_of_node p2) "uppercase identifier expected"
+            | _ -> error (loc_of_patt p2) "uppercase identifier expected"
             end
-        | _ -> error (loc_of_node p2) "bad pattern"
+        | _ -> error (loc_of_patt p2) "bad pattern"
       in
       mkpat loc p
   | PaAli (loc, p1, p2) ->
@@ -319,14 +340,23 @@ let rec patt =
           else
             let a = mkpat loc (Ppat_tuple al) in
             mkpat loc (Ppat_construct (li, Some a, true))
+      | Ppat_variant (s, None) ->
+          let a =
+            match al with
+              [a] -> a
+            | _ -> mkpat loc (Ppat_tuple al)
+          in
+          mkpat loc (Ppat_variant (s, Some a))
       | _ ->
-          error (loc_of_node f)
+          error (loc_of_patt f)
             "this is not a constructor, it cannot be applied in a pattern"
       end
   | PaArr (loc, pl) -> mkpat loc (Ppat_array (List.map patt pl))
   | PaChr (loc, s) -> mkpat loc (Ppat_constant (Const_char s))
   | PaInt (loc, s) -> mkpat loc (Ppat_constant (Const_int (int_of_string s)))
+  | PaLab (loc, _, _) -> error loc "labeled pattern not allowed here"
   | PaLid (loc, s) -> mkpat loc (Ppat_var (shd s))
+  | PaOlb (loc, _, _, _) -> error loc "labeled pattern not allowed here"
   | PaOrp (loc, p1, p2) -> mkpat loc (Ppat_or (patt p1, patt p2))
   | PaRng (loc, p1, p2) ->
       begin match p1, p2 with
@@ -339,6 +369,7 @@ let rec patt =
   | PaTyc (loc, p, t) -> mkpat loc (Ppat_constraint (patt p, ctyp t))
   | PaUid (loc, s) ->
       mkpat loc (Ppat_construct (lident (conv_con s), None, true))
+  | PaVrn (loc, s) -> mkpat loc (Ppat_variant (s, None))
   | PaXnd (loc, c, _) ->
       error loc ("pattern \"" ^ c ^ "\" (extension) not allowed here")
 and mklabpat (lab, p) = patt_label_long_id lab, patt p;;
@@ -357,13 +388,13 @@ let rec sep_expr_acc l =
         [] -> [loc, [], e]
       | ((_, ep), sl, e) :: l -> ((bp, ep), s :: sl, e) :: l
       end
-  | e -> (loc_of_node e, [], e) :: l
+  | e -> (loc_of_expr e, [], e) :: l
 ;;
 
 let expr_label_long_id e =
   match sep_expr_acc [] e with
     [_, ml, ExLid (_, s)] -> mkli (conv_lab s) ml
-  | _ -> error (loc_of_node e) "invalid label"
+  | _ -> error (loc_of_expr e) "invalid label"
 ;;
 
 let class_info class_expr ci =
@@ -391,16 +422,17 @@ let rec expr =
                ExLid (_, s) ->
                  let loc = bp, ep in
                  loc, mkexp loc (Pexp_field (e1, mkli (conv_lab s) ml))
-             | _ -> error (loc_of_node e2) "lowercase identifier expected")
+             | _ -> error (loc_of_expr e2) "lowercase identifier expected")
           (loc, e) l
       in
       e
   | ExAnt (_, e) -> expr e
   | ExApp (loc, _, _) as f ->
       let (f, al) = expr_fa [] f in
-      let al = List.map expr al in
+      let al = List.map label_expr al in
       begin match (expr f).pexp_desc with
         Pexp_construct (li, None, _) ->
+          let al = List.map snd al in
           if !no_constructors_arity then
             let a =
               match al with
@@ -411,13 +443,21 @@ let rec expr =
           else
             let a = mkexp loc (Pexp_tuple al) in
             mkexp loc (Pexp_construct (li, Some a, true))
+      | Pexp_variant (s, None) ->
+          let al = List.map snd al in
+          let a =
+            match al with
+              [a] -> a
+            | _ -> mkexp loc (Pexp_tuple al)
+          in
+          mkexp loc (Pexp_variant (s, Some a))
       | _ -> mkexp loc (Pexp_apply (expr f, al))
       end
   | ExAre (loc, e1, e2) ->
       mkexp loc
         (Pexp_apply
            (mkexp loc (Pexp_ident (array_function "Array" "get")),
-            [expr e1; expr e2]))
+            ["", expr e1; "", expr e2]))
   | ExArr (loc, el) -> mkexp loc (Pexp_array (List.map expr el))
   | ExAss (loc, e, v) ->
       let e =
@@ -430,12 +470,12 @@ let rec expr =
         | ExAre (_, e1, e2) ->
             Pexp_apply
               (mkexp loc (Pexp_ident (array_function "Array" "set")),
-               [expr e1; expr e2; expr v])
+               ["", expr e1; "", expr e2; "", expr v])
         | ExLid (_, lab) -> Pexp_setinstvar (shd lab, expr v)
         | ExSte (_, e1, e2) ->
             Pexp_apply
               (mkexp loc (Pexp_ident (array_function "String" "set")),
-               [expr e1; expr e2; expr v])
+               ["", expr e1; "", expr e2; "", expr v])
         | _ -> error loc "bad left part of assignment"
       in
       mkexp loc e
@@ -456,10 +496,17 @@ let rec expr =
       in
       let df = if df then Upto else Downto in
       mkexp loc (Pexp_for (shd i, expr e1, expr e2, df, expr e3))
-  | ExFun (loc, pel) -> mkexp loc (Pexp_function (List.map mkpwe pel))
+  | ExFun (loc, [PaLab (_, lab, p), w, e]) ->
+      mkexp loc (Pexp_function (lab, None, [patt p, when_expr e w]))
+  | ExFun (loc, [PaOlb (_, lab, p, eo), w, e]) ->
+      mkexp loc
+        (Pexp_function (("?" ^ lab), option expr eo, [patt p, when_expr e w]))
+  | ExFun (loc, pel) ->
+      mkexp loc (Pexp_function ("", None, List.map mkpwe pel))
   | ExIfe (loc, e1, e2, e3) ->
       mkexp loc (Pexp_ifthenelse (expr e1, expr e2, Some (expr e3)))
   | ExInt (loc, s) -> mkexp loc (Pexp_constant (Const_int (int_of_string s)))
+  | ExLab (loc, _, _) -> error loc "labeled expression not allowed here"
   | ExLet (loc, rf, pel, e) ->
       mkexp loc (Pexp_let (mkrf rf, List.map mkpe pel, expr e))
   | ExLid (loc, s) -> mkexp loc (Pexp_ident (lident s))
@@ -467,6 +514,7 @@ let rec expr =
       mkexp loc (Pexp_letmodule (i, module_expr me, expr e))
   | ExMat (loc, e, pel) -> mkexp loc (Pexp_match (expr e, List.map mkpwe pel))
   | ExNew (loc, id) -> mkexp loc (Pexp_new (long_id_of_string_list loc id))
+  | ExOlb (loc, _, _) -> error loc "labeled expression not allowed here"
   | ExOvr (loc, iel) -> mkexp loc (Pexp_override (List.map mkideexp iel))
   | ExRec (loc, lel, eo) ->
       let eo =
@@ -483,7 +531,7 @@ let rec expr =
       mkexp loc
         (Pexp_apply
            (mkexp loc (Pexp_ident (array_function "String" "get")),
-            [expr e1; expr e2]))
+            ["", expr e1; "", expr e2]))
   | ExStr (loc, s) -> mkexp loc (Pexp_constant (Const_string (shd s)))
   | ExTry (loc, e, pel) -> mkexp loc (Pexp_try (expr e, List.map mkpwe pel))
   | ExTup (loc, el) -> mkexp loc (Pexp_tuple (List.map expr el))
@@ -491,6 +539,7 @@ let rec expr =
       mkexp loc (Pexp_constraint (expr e, Some (ctyp t), None))
   | ExUid (loc, s) ->
       mkexp loc (Pexp_construct (lident (conv_con s), None, true))
+  | ExVrn (loc, s) -> mkexp loc (Pexp_variant (s, None))
   | ExWhi (loc, e1, el) ->
       let e2 =
         match List.rev el with
@@ -500,11 +549,17 @@ let rec expr =
       mkexp loc (Pexp_while (expr e1, expr e2))
   | ExXnd (loc, c, _) ->
       error loc ("expression \"" ^ c ^ "\" (extension) not allowed here")
+and label_expr =
+  function
+    ExLab (loc, lab, e) -> lab, expr e
+  | ExOlb (loc, lab, e) -> "?" ^ lab, expr e
+  | e -> "", expr e
 and mkpe (p, e) = patt p, expr e
-and mkpwe (p, w, e) =
-  match w with
-    Some w -> patt p, mkexp (loc_of_node e) (Pexp_when (expr w, expr e))
-  | _ -> patt p, expr e
+and mkpwe (p, w, e) = patt p, when_expr e w
+and when_expr e =
+  function
+    Some w -> mkexp (loc_of_expr e) (Pexp_when (expr w, expr e))
+  | None -> expr e
 and mklabexp (lab, e) = expr_label_long_id lab, expr e
 and mkideexp (ide, e) = shd ide, expr e
 and mktype_decl (c, tl, td, cl) =
@@ -585,7 +640,12 @@ and class_type =
     CtCon (loc, id, tl) ->
       mkcty loc
         (Pcty_constr (long_id_of_string_list loc id, List.map ctyp tl))
-  | CtFun (loc, t, ct) -> mkcty loc (Pcty_fun (ctyp t, class_type ct))
+  | CtFun (loc, TyLab (_, lab, t), ct) ->
+      mkcty loc (Pcty_fun (lab, ctyp t, class_type ct))
+  | CtFun (loc, TyOlb (loc1, lab, t), ct) ->
+      let t = TyApp (loc1, TyLid (loc1, "option"), t) in
+      mkcty loc (Pcty_fun (("?" ^ lab), ctyp t, class_type ct))
+  | CtFun (loc, t, ct) -> mkcty loc (Pcty_fun ("", ctyp t, class_type ct))
   | CtSig (loc, t_o, ctfl) ->
       let t =
         match t_o with
@@ -606,10 +666,21 @@ and class_sig_item =
 and class_expr =
   function
     CeApp (loc, ce, el) ->
-      mkpcl loc (Pcl_apply (class_expr ce, List.map expr el))
+      let el = List.map label_expr el in
+      mkpcl loc (Pcl_apply (class_expr ce, el))
   | CeCon (loc, id, tl) ->
       mkpcl loc (Pcl_constr (long_id_of_string_list loc id, List.map ctyp tl))
-  | CeFun (loc, p, ce) -> mkpcl loc (Pcl_fun (patt p, class_expr ce))
+  | CeFun (loc, PaLab (_, lab, p), ce) ->
+      mkpcl loc (Pcl_fun (lab, None, patt p, class_expr ce))
+  | CeFun (loc, PaOlb (_, lab, p, eo), ce) ->
+      let eo =
+        match eo with
+          Some e -> Some (expr e)
+        | None -> None
+      in
+      mkpcl loc (Pcl_fun (("?" ^ lab), eo, patt p, class_expr ce))
+  | CeFun (loc, p, ce) ->
+      mkpcl loc (Pcl_fun ("", None, patt p, class_expr ce))
   | CeLet (loc, rf, pel, ce) ->
       mkpcl loc (Pcl_let (mkrf rf, List.map mkpe pel, class_expr ce))
   | CeStr (loc, po, cfl) ->
