@@ -309,6 +309,46 @@ let recover parser_of_tree entry nlevn alevn bp a s son strm =
   else do_recover parser_of_tree entry nlevn alevn bp a s son strm
 ;;
 
+let std_token_parse =
+  function
+    p_con, "" ->
+      (fun (strm__ : _ Stream.t) ->
+         match Stream.peek strm__ with
+           Some (con, prm) when con = p_con -> Stream.junk strm__; prm
+         | _ -> raise Stream.Failure)
+  | p_con, p_prm ->
+      fun (strm__ : _ Stream.t) ->
+        match Stream.peek strm__ with
+          Some (con, prm) when con = p_con && prm = p_prm ->
+            Stream.junk strm__; prm
+        | _ -> raise Stream.Failure
+;;
+
+let peek_nth n strm =
+  let list = Stream.npeek n strm in
+  let rec loop list n =
+    match list, n with
+      x :: _, 1 -> Some x
+    | _ :: l, n -> loop l (n - 1)
+    | [], _ -> None
+  in
+  loop list n
+;;
+
+let rec get_token_list entry tokl last_tok tree =
+  match tree with
+    Node {node = Stoken tok as s; son = son; brother = DeadEnd} ->
+      begin match entry.egram.glexer.Token.tparse tok with
+        Some _ ->
+          if tokl = [] then None
+          else Some (List.rev (last_tok :: tokl), last_tok, tree)
+      | None -> get_token_list entry (last_tok :: tokl) tok son
+      end
+  | _ ->
+      if tokl = [] then None
+      else Some (List.rev (last_tok :: tokl), last_tok, tree)
+;;
+
 let rec parser_of_tree entry nlevn alevn =
   function
     DeadEnd -> (fun (strm__ : _ Stream.t) -> raise Stream.Failure)
@@ -326,55 +366,123 @@ let rec parser_of_tree entry nlevn alevn =
            Some a -> app act a
          | _ -> p2 strm__)
   | Node {node = s; son = son; brother = DeadEnd} ->
-      let ps = parser_of_symbol entry nlevn s in
-      let p1 = parser_of_tree entry nlevn alevn son in
-      (fun (strm__ : _ Stream.t) ->
-         let bp = Stream.count strm__ in
-         let a = ps strm__ in
-         let act =
-           try
-             (fun (strm__ : _ Stream.t) ->
-                try p1 strm__ with
-                  Stream.Failure ->
-                    try
-                      recover parser_of_tree entry nlevn alevn bp a s son
-                        strm__
-                    with
-                      Stream.Failure ->
-                        raise (Stream.Error (tree_failed entry a s son)))
-               strm__
-           with
-             Stream.Failure -> raise (Stream.Error "")
-         in
-         app act a)
+      let tokl =
+        match s with
+          Stoken tok when entry.egram.glexer.Token.tparse tok = None ->
+            get_token_list entry [] tok son
+        | _ -> None
+      in
+      begin match tokl with
+        None ->
+          let ps = parser_of_symbol entry nlevn s in
+          let p1 = parser_of_tree entry nlevn alevn son in
+          let p1 = parser_cont p1 entry nlevn alevn s son in
+          (fun (strm__ : _ Stream.t) ->
+             let bp = Stream.count strm__ in
+             let a = ps strm__ in
+             let act =
+               try p1 bp a strm__ with
+                 Stream.Failure -> raise (Stream.Error "")
+             in
+             app act a)
+      | Some (tokl, last_tok, son) ->
+          let p1 = parser_of_tree entry nlevn alevn son in
+          let p1 = parser_cont p1 entry nlevn alevn (Stoken last_tok) son in
+          parser_of_token_list p1 tokl
+      end
   | Node {node = s; son = son; brother = bro} ->
-      let ps = parser_of_symbol entry nlevn s in
-      let p1 = parser_of_tree entry nlevn alevn son in
-      let p2 = parser_of_tree entry nlevn alevn bro in
-      fun (strm__ : _ Stream.t) ->
-        let bp = Stream.count strm__ in
-        match
-          try Some (ps strm__) with
-            Stream.Failure -> None
-        with
-          Some a ->
-            let act =
-              try
-                (fun (strm__ : _ Stream.t) ->
-                   try p1 strm__ with
-                     Stream.Failure ->
-                       try
-                         recover parser_of_tree entry nlevn alevn bp a s son
-                           strm__
-                       with
-                         Stream.Failure ->
-                           raise (Stream.Error (tree_failed entry a s son)))
-                  strm__
-              with
-                Stream.Failure -> raise (Stream.Error "")
-            in
-            app act a
-        | _ -> p2 strm__
+      let tokl =
+        match s with
+          Stoken tok when entry.egram.glexer.Token.tparse tok = None ->
+            get_token_list entry [] tok son
+        | _ -> None
+      in
+      match tokl with
+        None ->
+          let ps = parser_of_symbol entry nlevn s in
+          let p1 = parser_of_tree entry nlevn alevn son in
+          let p1 = parser_cont p1 entry nlevn alevn s son in
+          let p2 = parser_of_tree entry nlevn alevn bro in
+          (fun (strm__ : _ Stream.t) ->
+             let bp = Stream.count strm__ in
+             match
+               try Some (ps strm__) with
+                 Stream.Failure -> None
+             with
+               Some a ->
+                 let act =
+                   try p1 bp a strm__ with
+                     Stream.Failure -> raise (Stream.Error "")
+                 in
+                 app act a
+             | _ -> p2 strm__)
+      | Some (tokl, last_tok, son) ->
+          let p1 = parser_of_tree entry nlevn alevn son in
+          let p1 = parser_cont p1 entry nlevn alevn (Stoken last_tok) son in
+          let p1 = parser_of_token_list p1 tokl in
+          let p2 = parser_of_tree entry nlevn alevn bro in
+          fun (strm__ : _ Stream.t) ->
+            try p1 strm__ with
+              Stream.Failure -> p2 strm__
+and parser_cont p1 entry nlevn alevn s son bp a (strm__ : _ Stream.t) =
+  try p1 strm__ with
+    Stream.Failure ->
+      try recover parser_of_tree entry nlevn alevn bp a s son strm__ with
+        Stream.Failure -> raise (Stream.Error (tree_failed entry a s son))
+and parser_of_token_list p1 tokl =
+  let rec loop n =
+    function
+      [p_con, ""] ->
+        let ps (strm__ : _ Stream.t) =
+          match peek_nth n strm__ with
+            Some (con, prm) when con = p_con ->
+              for i = 1 to n do Stream.junk strm__ done; Obj.repr prm
+          | _ -> raise Stream.Failure
+        in
+        (fun (strm__ : _ Stream.t) ->
+           let bp = Stream.count strm__ in
+           let a = ps strm__ in
+           let act =
+             try p1 bp a strm__ with
+               Stream.Failure -> raise (Stream.Error "")
+           in
+           app act a)
+    | [p_con, p_prm] ->
+        let ps (strm__ : _ Stream.t) =
+          match peek_nth n strm__ with
+            Some (con, prm) when con = p_con && prm = p_prm ->
+              for i = 1 to n do Stream.junk strm__ done; Obj.repr prm
+          | _ -> raise Stream.Failure
+        in
+        (fun (strm__ : _ Stream.t) ->
+           let bp = Stream.count strm__ in
+           let a = ps strm__ in
+           let act =
+             try p1 bp a strm__ with
+               Stream.Failure -> raise (Stream.Error "")
+           in
+           app act a)
+    | (p_con, "") :: tokl ->
+        let ps (strm__ : _ Stream.t) =
+          match peek_nth n strm__ with
+            Some (con, prm) when con = p_con -> prm
+          | _ -> raise Stream.Failure
+        in
+        let p1 = loop (n + 1) tokl in
+        (fun (strm__ : _ Stream.t) ->
+           let a = ps strm__ in let act = p1 strm__ in app act a)
+    | (p_con, p_prm) :: tokl ->
+        let ps (strm__ : _ Stream.t) =
+          match peek_nth n strm__ with
+            Some (con, prm) when con = p_con && prm = p_prm -> prm
+          | _ -> raise Stream.Failure
+        in
+        let p1 = loop (n + 1) tokl in
+        (fun (strm__ : _ Stream.t) ->
+           let a = ps strm__ in let act = p1 strm__ in app act a)
+    | [] -> assert false
+  in
+  loop 1 tokl
 and parser_of_symbol entry nlevn =
   function
     Slist0 s ->
@@ -435,20 +543,13 @@ and parser_of_symbol entry nlevn =
         with
           Some v ->
             let a =
-              try
-                (fun (strm__ : _ Stream.t) ->
-                   try ps strm__ with
-                     Stream.Failure ->
-                       try
-                         parser_of_symbol entry nlevn (top_symb entry symb)
-                           strm__
-                       with
-                         Stream.Failure ->
-                           raise
-                             (Stream.Error (symb_failed entry v sep symb)))
-                  strm__
-              with
-                Stream.Failure -> raise (Stream.Error "")
+              try ps strm__ with
+                Stream.Failure ->
+                  try
+                    parser_of_symbol entry nlevn (top_symb entry symb) strm__
+                  with
+                    Stream.Failure ->
+                      raise (Stream.Error (symb_failed entry v sep symb))
             in
             kont (a :: al) strm__
         | _ -> al
@@ -477,9 +578,9 @@ and parser_of_symbol entry nlevn =
   | Sself -> (fun (strm__ : _ Stream.t) -> entry.estart 0 strm__)
   | Snext -> (fun (strm__ : _ Stream.t) -> entry.estart nlevn strm__)
   | Stoken tok ->
-      (Obj.magic
-         (entry.egram.glexer.Token.tparse tok : Token.t Stream.t -> string) :
-       Token.t Stream.t -> Obj.t)
+      match entry.egram.glexer.Token.tparse tok with
+        Some f -> (Obj.magic f : Token.t Stream.t -> Obj.t)
+      | None -> (Obj.magic (std_token_parse tok) : Token.t Stream.t -> Obj.t)
 ;;
 
 let rec continue_parser_of_levels entry clevn =
