@@ -164,7 +164,7 @@ let rec skip_indent (strm__ : _ Stream.t) =
   | _ -> ()
 ;;
 
-let skip_opt_newline (strm__ : _ Stream.t) =
+let skip_opt_linefeed (strm__ : _ Stream.t) =
   match Stream.peek strm__ with
     Some '\n' -> Stream.junk strm__; ()
   | _ -> ()
@@ -209,23 +209,69 @@ let next_token_fun find_id_kwd find_spe_kwd =
         end
     | Some '"' -> Stream.junk strm__; "STRING", string bp 0 strm__
     | Some '$' -> Stream.junk strm__; locate_or_antiquot bp 0 strm__
-    | Some
-        ('!' | '?' | '~' | '=' | '@' | '^' | '&' | '+' | '-' | '*' | '/' | '%'
-           as c) ->
+    | Some ('!' | '=' | '@' | '^' | '&' | '+' | '-' | '*' | '/' | '%' as c) ->
         Stream.junk strm__;
         let id = get_buff (ident2 (store 0 c) strm__) in
         keyword_or_error (bp, Stream.count strm__) id
+    | Some ('~' as c) ->
+        Stream.junk strm__;
+        begin try
+          match Stream.peek strm__ with
+            Some ('a'..'z' as c) ->
+              Stream.junk strm__;
+              let len =
+                try ident (store 0 c) strm__ with
+                  Stream.Failure -> raise (Stream.Error "")
+              in
+              let t =
+                try
+                  match Stream.peek strm__ with
+                    Some ':' -> Stream.junk strm__; "TILDEIDENTCOLON"
+                  | _ -> "TILDEIDENT"
+                with
+                  Stream.Failure -> raise (Stream.Error "")
+              in
+              t, get_buff len
+          | _ ->
+              let id = get_buff (ident2 (store 0 c) strm__) in
+              keyword_or_error (bp, Stream.count strm__) id
+        with
+          Stream.Failure -> raise (Stream.Error "")
+        end
+    | Some ('?' as c) ->
+        Stream.junk strm__;
+        begin try
+          match Stream.peek strm__ with
+            Some ('a'..'z' as c) ->
+              Stream.junk strm__;
+              let len =
+                try ident (store 0 c) strm__ with
+                  Stream.Failure -> raise (Stream.Error "")
+              in
+              let t =
+                try
+                  match Stream.peek strm__ with
+                    Some ':' -> Stream.junk strm__; "QUESTIONIDENTCOLON"
+                  | _ -> "QUESTIONIDENT"
+                with
+                  Stream.Failure -> raise (Stream.Error "")
+              in
+              t, get_buff len
+          | _ ->
+              let id = get_buff (ident2 (store 0 c) strm__) in
+              keyword_or_error (bp, Stream.count strm__) id
+        with
+          Stream.Failure -> raise (Stream.Error "")
+        end
     | Some '<' -> Stream.junk strm__; less bp strm__
     | Some (':' as c1) ->
         Stream.junk strm__;
         let len =
           try
-            (fun (strm__ : _ Stream.t) ->
-               match Stream.peek strm__ with
-                 Some (']' | ':' | '=' | '>' as c2) ->
-                   Stream.junk strm__; store (store 0 c1) c2
-               | _ -> store 0 c1)
-              strm__
+            match Stream.peek strm__ with
+              Some (']' | ':' | '=' | '>' as c2) ->
+                Stream.junk strm__; store (store 0 c1) c2
+            | _ -> store 0 c1
           with
             Stream.Failure -> raise (Stream.Error "")
         in
@@ -235,12 +281,10 @@ let next_token_fun find_id_kwd find_spe_kwd =
         Stream.junk strm__;
         let len =
           try
-            (fun (strm__ : _ Stream.t) ->
-               match Stream.peek strm__ with
-                 Some (']' | '}' as c2) ->
-                   Stream.junk strm__; store (store 0 c1) c2
-               | _ -> ident2 (store 0 c1) strm__)
-              strm__
+            match Stream.peek strm__ with
+              Some (']' | '}' as c2) ->
+                Stream.junk strm__; store (store 0 c1) c2
+            | _ -> ident2 (store 0 c1) strm__
           with
             Stream.Failure -> raise (Stream.Error "")
         in
@@ -265,11 +309,9 @@ let next_token_fun find_id_kwd find_spe_kwd =
         Stream.junk strm__;
         let id =
           try
-            (fun (strm__ : _ Stream.t) ->
-               match Stream.peek strm__ with
-                 Some '.' -> Stream.junk strm__; ".."
-               | _ -> ".")
-              strm__
+            match Stream.peek strm__ with
+              Some '.' -> Stream.junk strm__; ".."
+            | _ -> "."
           with
             Stream.Failure -> raise (Stream.Error "")
         in
@@ -278,11 +320,9 @@ let next_token_fun find_id_kwd find_spe_kwd =
         Stream.junk strm__;
         let id =
           try
-            (fun (strm__ : _ Stream.t) ->
-               match Stream.peek strm__ with
-                 Some ';' -> Stream.junk strm__; ";;"
-               | _ -> ";")
-              strm__
+            match Stream.peek strm__ with
+              Some ';' -> Stream.junk strm__; ";;"
+            | _ -> ";"
           with
             Stream.Failure -> raise (Stream.Error "")
         in
@@ -301,11 +341,7 @@ let next_token_fun find_id_kwd find_spe_kwd =
     | Some ':' ->
         Stream.junk strm__;
         let i =
-          try
-            (fun (strm__ : _ Stream.t) ->
-               let len = ident 0 strm__ in get_buff len)
-              strm__
-          with
+          try let len = ident 0 strm__ in get_buff len with
             Stream.Failure -> raise (Stream.Error "")
         in
         begin match Stream.peek strm__ with
@@ -325,38 +361,35 @@ let next_token_fun find_id_kwd find_spe_kwd =
         Stream.junk strm__;
         let len =
           try
-            (fun (strm__ : _ Stream.t) ->
-               match Stream.peek strm__ with
-                 Some '\n' ->
-                   Stream.junk strm__;
-                   let _ =
-                     try skip_indent strm__ with
-                       Stream.Failure -> raise (Stream.Error "")
-                   in
-                   len
-               | Some '\013' ->
-                   Stream.junk strm__;
-                   let _ =
-                     try skip_opt_newline strm__ with
-                       Stream.Failure -> raise (Stream.Error "")
-                   in
-                   let _ =
-                     try skip_indent strm__ with
-                       Stream.Failure -> raise (Stream.Error "")
-                   in
-                   len
-               | _ ->
-                   match
-                     try Some (escape_string strm__) with
-                       Stream.Failure -> None
-                   with
-                     Some c -> store len c
-                   | _ ->
-                       match Stream.peek strm__ with
-                         Some c ->
-                           Stream.junk strm__; store (store len '\\') c
-                       | _ -> raise Stream.Failure)
-              strm__
+            match Stream.peek strm__ with
+              Some '\n' ->
+                Stream.junk strm__;
+                let _ =
+                  try skip_indent strm__ with
+                    Stream.Failure -> raise (Stream.Error "")
+                in
+                len
+            | Some '\013' ->
+                Stream.junk strm__;
+                let _ =
+                  try skip_opt_linefeed strm__ with
+                    Stream.Failure -> raise (Stream.Error "")
+                in
+                let _ =
+                  try skip_indent strm__ with
+                    Stream.Failure -> raise (Stream.Error "")
+                in
+                len
+            | _ ->
+                match
+                  try Some (escape_string strm__) with
+                    Stream.Failure -> None
+                with
+                  Some c -> store len c
+                | _ ->
+                    match Stream.peek strm__ with
+                      Some c -> Stream.junk strm__; store (store len '\\') c
+                    | _ -> raise Stream.Failure
           with
             Stream.Failure -> raise (Stream.Error "")
         in
@@ -478,12 +511,9 @@ let next_token_fun find_id_kwd find_spe_kwd =
         Stream.junk strm__;
         let len =
           try
-            (fun (strm__ : _ Stream.t) ->
-               match Stream.peek strm__ with
-                 Some ('>' | '<' | '\\' as c) ->
-                   Stream.junk strm__; store len c
-               | _ -> store len '\\')
-              strm__
+            match Stream.peek strm__ with
+              Some ('>' | '<' | '\\' as c) -> Stream.junk strm__; store len c
+            | _ -> store len '\\'
           with
             Stream.Failure -> raise (Stream.Error "")
         in
@@ -502,13 +532,11 @@ let next_token_fun find_id_kwd find_spe_kwd =
             Stream.Failure -> raise (Stream.Error "")
         in
         begin try
-          (fun (strm__ : _ Stream.t) ->
-             match Stream.peek strm__ with
-               Some '<' ->
-                 Stream.junk strm__;
-                 mstore (quotation bp (store len '<') strm__) ">>"
-             | _ -> len)
-            strm__
+          match Stream.peek strm__ with
+            Some '<' ->
+              Stream.junk strm__;
+              mstore (quotation bp (store len '<') strm__) ">>"
+          | _ -> len
         with
           Stream.Failure -> raise (Stream.Error "")
         end
@@ -668,11 +696,9 @@ and check (strm__ : _ Stream.t) =
       Stream.junk strm__;
       let _ =
         try
-          (fun (strm__ : _ Stream.t) ->
-             match Stream.peek strm__ with
-               Some (']' | ':' | '=' | '>') -> Stream.junk strm__; ()
-             | _ -> ())
-            strm__
+          match Stream.peek strm__ with
+            Some (']' | ':' | '=' | '>') -> Stream.junk strm__; ()
+          | _ -> ()
         with
           Stream.Failure -> raise (Stream.Error "")
       in
@@ -681,11 +707,9 @@ and check (strm__ : _ Stream.t) =
       Stream.junk strm__;
       let _ =
         try
-          (fun (strm__ : _ Stream.t) ->
-             match Stream.peek strm__ with
-               Some (']' | '}') -> Stream.junk strm__; ()
-             | _ -> check_ident2 strm__)
-            strm__
+          match Stream.peek strm__ with
+            Some (']' | '}') -> Stream.junk strm__; ()
+          | _ -> check_ident2 strm__
         with
           Stream.Failure -> raise (Stream.Error "")
       in
@@ -705,11 +729,9 @@ and check (strm__ : _ Stream.t) =
       Stream.junk strm__;
       let _ =
         try
-          (fun (strm__ : _ Stream.t) ->
-             match Stream.peek strm__ with
-               Some ';' -> Stream.junk strm__; ()
-             | _ -> ())
-            strm__
+          match Stream.peek strm__ with
+            Some ';' -> Stream.junk strm__; ()
+          | _ -> ()
         with
           Stream.Failure -> raise (Stream.Error "")
       in
@@ -749,7 +771,8 @@ let using_token kwd_table (p_con, p_prm) =
                  ("the token \"" ^ p_prm ^
                     "\" does not respect Plexer rules"))
       end
-  | "LIDENT" | "UIDENT" | "INT" | "FLOAT" | "CHAR" | "STRING" | "QUOTATION" |
+  | "LIDENT" | "UIDENT" | "TILDEIDENT" | "TILDEIDENTCOLON" | "QUESTIONIDENT" |
+    "QUESTIONIDENTCOLON" | "INT" | "FLOAT" | "CHAR" | "STRING" | "QUOTATION" |
     "ANTIQUOT" | "LOCATE" | "EOI" ->
       ()
   | _ ->
@@ -804,22 +827,14 @@ let after_colon e =
 let tparse =
   function
     "ANTIQUOT", p_prm ->
-      (fun (strm__ : _ Stream.t) ->
-         match Stream.peek strm__ with
-           Some ("ANTIQUOT", prm) when eq_before_colon p_prm prm ->
-             Stream.junk strm__; after_colon prm
-         | _ -> raise Stream.Failure)
-  | p_con, "" ->
-      (fun (strm__ : _ Stream.t) ->
-         match Stream.peek strm__ with
-           Some (con, prm) when con = p_con -> Stream.junk strm__; prm
-         | _ -> raise Stream.Failure)
-  | p_con, p_prm ->
-      fun (strm__ : _ Stream.t) ->
+      let p (strm__ : _ Stream.t) =
         match Stream.peek strm__ with
-          Some (con, prm) when con = p_con && prm = p_prm ->
-            Stream.junk strm__; prm
+          Some ("ANTIQUOT", prm) when eq_before_colon p_prm prm ->
+            Stream.junk strm__; after_colon prm
         | _ -> raise Stream.Failure
+      in
+      Some p
+  | _ -> None
 ;;
 
 let make () =
