@@ -23,6 +23,6 @@ type lexer =
   { func : char Stream.t -> t Stream.t * location_function;
     using : pattern -> unit;
     removing : pattern -> unit;
-    tparse : pattern -> t Stream.t -> string;
+    tparse : pattern -> (t Stream.t -> string) option;
     text : pattern -> string }
 ;;
