@@ -37,7 +37,7 @@ type lexer =
   { func : char Stream.t -> t Stream.t * location_function;
     using : pattern -> unit;
     removing : pattern -> unit;
-    tparse : pattern -> t Stream.t -> string;
+    tparse : pattern -> (t Stream.t -> string) option;
     text : pattern -> string }
 ;;
     (* The type for a lexer used by Camlp4 grammars.
@@ -52,7 +52,8 @@ type lexer =
        does not uses the given token (pattern) any more. If the lexer has a
        notion of "keywords", it can release it from its tables. Called by
        [DELETE_RULE] statements.
--      The field [tparse] is a function returning the parsing function
-       associated with each token pattern.
+-      The field [tparse] is a function returning either [Some f], f being
+       the parsing function associated with the pattern, or [None] if it
+       is a standard parsing of the pattern.
 -      The field [text] returns the name of some token pattern,
        used in error messages. *)
