@@ -132,7 +132,17 @@ let stream_pattern_component skont ckont =
   | SpNtr (loc, p, e) ->
       let e =
         match e with
-          MLast.ExFun (_, [MLast.PaLid (_, v), None, e]) when v = strm_n -> e
+          MLast.ExFun
+            (_,
+             [MLast.PaTyc
+                (_, MLast.PaLid (_, v),
+                 MLast.TyApp
+                   (_,
+                    MLast.TyAcc
+                      (_, MLast.TyUid (_, "Stream"), MLast.TyLid (_, "t")),
+                    MLast.TyAny _)), None, e])
+          when v = strm_n ->
+            e
         | _ -> MLast.ExApp (loc, e, MLast.ExLid (loc, strm_n))
       in
       if pattern_eq_expression p skont then
